from pygeppetto.data_model import GeppettoProject
from pygeppetto.managers.runtime_experiment import RuntimeExperiment
from pygeppetto.model import GeppettoModel, model_utility, CompoundRefQuery, DataSource
from pygeppetto.model.model_access import GeppettoModelAccess
from pygeppetto.model.types import ImportType
from pygeppetto.model.utils import model_traversal
from pygeppetto.model.utils import url_reader
from pygeppetto.services import model_interpreter

from pygeppetto.model.exceptions import GeppettoExecutionException, GeppettoModelException
from pygeppetto.services.data_source_service import ServiceCreator, DataSourceService


class RuntimeProject(object):
    def __init__(self, project: GeppettoProject, manager):
        """
        :param project: pygeppetto.data_model.GeppettoProject
        :param manager: GeppettoManager
        """

        self.geppetto_manager = manager
        self.project = project
        self.experiments = {}
        self.active_experiment = None
        if isinstance(project.geppettoModel, GeppettoModel):
            self.model = project.geppettoModel
        elif hasattr(project.geppettoModel, 'url'):
            raise NotImplementedError('PersistedData in project is not supported: cannot load model')
        else:
            raise GeppettoModelException('Bad formed project: model is not correctly specified')

        self.geppetto_model_access = GeppettoModelAccess(self.model)
        self.import_autoresolve_types(self.model)
        self.data_source_services = {}

        # TODO handle views
        # TODO handle experiments

    def release(self):
        pass

    def import_autoresolve_types(self, model):
        model_traversal.apply_single(model, self.import_type,
                                     lambda node: type(node) == ImportType and node.autoresolve)

    def import_type(self, itype: ImportType):
        library = itype.eContainer()

        url = itype.url if None is None else url_reader.getURL(itype.url, self.project.base_url)
        actual_model_interpreter = model_interpreter.get_model_interpreter_from_library(library)
        newtype = actual_model_interpreter.importType(url, itype.id, library, self.geppetto_model_access)
        self.geppetto_model_access.swap_type(itype, newtype)

    def open_experiment(self, experiment):
        self.experiments[experiment] = RuntimeExperiment(self, experiment)
        self.active_experiment = experiment

    def get_runtime_experiment(self, experiment):
        return self.experiments[experiment]

    def __getitem__(self, item):
        return self.get_runtime_experiment(item)

    def resolve_import_value(self, path):
        """
        Retrieves the value from the path and updates the Geppetto Model with the new value
        :param path: the instance path to replace
        :return:
        """
        variable = self.geppetto_model_access.get_variable(path)

        variable.synched = False  # This will say to the serializer to send the value
        variable_container = variable.eContainer()

        source_library = variable_container.eContainer()

        # here we are simplifying the logic to retrieve the model interpreter. In Java geppetto here we have a switch-visitors call, we don't need that here anyway, unless we're missing something important
        actual_model_interpreter = model_interpreter.get_model_interpreter_from_library(source_library)

        var_to_import = self.geppetto_model_access.get_variable(path)
        value = var_to_import.initialValues[0].value
        new_value = actual_model_interpreter.importValue(value)

        # Set the new value in replacement of ImportValue
        variable.initialValues[0].value = new_value
        # TODO? it would be nice here to have an api that sets a value and also handles the sync
        variable_container.synched = False
        source_library.synched = False
        return self.model

    def resolve_import_type(self, type_paths):
        """Loads types to the model"""

        #  let's find the importType

        for type_path in type_paths:
            itype = self.geppetto_model_access.get_type(type_path)

            try:
                self.import_type(itype)

                # TODO if we want to support default view customization, uncomment below and implement
                # if self.gatherDefaultView and actual_model_interpreter.isSupported(
                #         GeppettoFeature.DEFAULT_VIEW_CUSTOMISER_FEATURE):
                #     viewCustomisations.add((actual_model_interpreter.getFeature(
                #         GeppettoFeature.DEFAULT_VIEW_CUSTOMISER_FEATURE)).getDefaultViewCustomisation(
                #         importedType))
            except Exception as e:
                raise GeppettoExecutionException('Error importing type ' + type_path) from e

        return self.model

    def get_data_source_service(self, data_source: DataSource):
        if not data_source.id in self.data_source_services:
            ds_service = ServiceCreator.get_new_datasource_service_instance(data_source,
                                                                            self.geppetto_model_access)
            self.data_source_services[data_source.id] = ds_service

        return self.data_source_services[data_source.id]

    def populate_new_experiment(self, experiment):
        raise NotImplemented

    def fetch_variable(self, data_source_id, variable_ids):
        """
        Fetch a variable on the geppetto model.
        :return:
        """
        data_source_service = self.get_data_source_service_by_id(data_source_id)

        for variable_id in variable_ids:
            if not variable_id in set(v.id for v in self.model.variables):
                data_source_service.fetch_variable(variable_id)
        return self.model

    def fetch(self, data_source_id, variable_ids, instance_ids):
        """
        Fetch variables and instances on the geppetto model.
        :return:
        """
        data_source_service = self.get_data_source_service_by_id(data_source_id)

        for variable_id in variable_ids:
            if not variable_id in set(v.id for v in self.geppetto_model_access.get_variables()):
                data_source_service.fetch_variable(variable_id)

        for instance_id in instance_ids:
            if not instance_id in set(v.id for v in self.geppetto_model_access.get_instances()):
                data_source_service.fetch_instance(instance_id)
        return self.model

    def run_query(self, queries):
        query = model_utility.get_query(queries[0].queryPath, self.model)
        if isinstance(query, CompoundRefQuery):
            # Use the first query of the chain to have the datasource we want to start from
            query = queries[0].queryPath
        data_source = query.eContainer()
        while not isinstance(data_source, DataSource):
            data_source = data_source.eContainer()
            assert data_source is not None, 'Bad data source definition'
        data_source_service = self.get_data_source_service(data_source)
        return data_source_service.execute(queries)

    def get_data_source_service_by_id(self, data_source_id) -> DataSourceService:
        if not data_source_id in self.data_source_services:
            try:
                ds = next(ds for ds in self.model.dataSources if ds.id == data_source_id)
            except StopIteration:
                raise GeppettoModelException("The datasource service for " + data_source_id + " was not found")
            return self.get_data_source_service(ds)
        return self.data_source_services[data_source_id]





from pyecore.resources import global_registry
from .types import getEClassifier, eClassifiers
from .types import name, nsURI, nsPrefix, eClass
from .types import Type, VisualType, ImportType, CompositeType, PointerType, QuantityType, ParameterType, StateVariableType, DynamicsType, ArgumentType, ExpressionType, HTMLType, JSONType, TextType, URLType, PointType, ArrayType, CompositeVisualType, ConnectionType, SimpleType, ImageType, SimpleArrayType, MetadataType

from ..values import VisualValue, Argument, ArrayValue, Image, Point, VisualGroup, Text, Quantity, JSON, Dynamics, URL, Composite, Pointer, HTML, AArrayValue, Expression
from ..model import Tag, DomainModel
from ..variables import Variable

from . import types
from .. import model


__all__ = ['Type', 'VisualType', 'ImportType', 'CompositeType', 'PointerType', 'QuantityType', 'ParameterType', 'StateVariableType', 'DynamicsType', 'ArgumentType', 'ExpressionType',
           'HTMLType', 'JSONType', 'TextType', 'URLType', 'PointType', 'ArrayType', 'CompositeVisualType', 'ConnectionType', 'SimpleType', 'ImageType', 'SimpleArrayType', 'MetadataType']

eSubpackages = []
eSuperPackage = model
types.eSubpackages = eSubpackages
types.eSuperPackage = eSuperPackage


otherClassifiers = []

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for classif in eClassifiers.values():
    eClass.eClassifiers.append(classif.eClass)

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)

register_packages = [types] + eSubpackages
for pack in register_packages:
    global_registry[pack.nsURI] = pack



import json
from pyecore.utils import dispatch
from pygeppetto.model import Node, CompoundQuery, ProcessQuery, CompoundRefQuery, Variable, Instance
from pygeppetto.model.datasources import Query, SimpleQuery
from pygeppetto.model.model_access import GeppettoModelAccess
from pygeppetto.model.utils import model_traversal
from pygeppetto.visitors import Switch
from pygeppetto.model.exceptions import GeppettoDataSourceException, GeppettoVisitingException, \
    GeppettoInitializationException
from pygeppetto.model.datasources.datasources import QueryResults, QueryResult, DataSource
from pygeppetto.model.utils.datasource import query_check
from pygeppetto.model.utils import template
from pygeppetto.utils import stream_requests

ID = "ID"


class ExecuteQueryVisitor(Switch):

    def __init__(self,
                 geppetto_model_access: GeppettoModelAccess,
                 node_path=None,
                 types=(),
                 count_only=False,
                 processing_output_map=None):
        self.node_id = node_path
        self.types = types
        self.geppetto_model_access = geppetto_model_access
        self.count = count_only
        self.results = None
        self.processing_output_map = processing_output_map if processing_output_map else {}

    @dispatch
    def do_switch(self, query):
        raise NotImplemented

    @do_switch.register(CompoundQuery)
    def case_compound_query(self, query: CompoundQuery):
        if self.count and not query.runForCount:
            return None
        run_query_visitor = ExecuteQueryVisitor(self.geppetto_model_access, self.node_id, self.types,
                                                processing_output_map=self.processing_output_map)
        model_traversal.apply_direct_children_only(query, run_query_visitor)
        self.merge_results(run_query_visitor.results)

    @do_switch.register(ProcessQuery)
    def case_process_query(self, query: ProcessQuery):
        if self.count and not query.runForCount:
            return None
        if not query_check(query, self.types):
            return None

        from pygeppetto.services.data_source_service import ServiceCreator
        try:
            qp = ServiceCreator.get_new_service_instance(query.queryProcessorId)

            self.results = qp.process(query, self.get_datasource(query=query), self.node_id, self.results,
                                      self.geppetto_model_access)
            self.processing_output_map = qp.get_processing_output_map()
        except GeppettoDataSourceException as e:
            raise GeppettoVisitingException(f"Data source exception while running query {query.id}") from e
        except GeppettoInitializationException as e:
            raise GeppettoVisitingException(f"Initialization exception while running query {query.id}") from e

    @do_switch.register(CompoundRefQuery)
    def case_compound_query_ref(self, query: CompoundRefQuery):
        raise NotImplemented

    @do_switch.register(SimpleQuery)
    def case_simple_query(self, query: SimpleQuery):
        """ Ported from
        https://github.com/openworm/org.geppetto.datasources/blob/development/src/main/java/org/geppetto/datasources/ExecuteQueryVisitor.java
        """
        if not self.count or (self.count and query.runForCount):
            try:
                if query_check(query, self.types):
                    # had to import here to avoid circular import error
                    from pygeppetto.services.data_source_service import ServiceCreator
                    ds = self.get_datasource(query=query)
                    dss = ServiceCreator.get_new_datasource_service_instance(data_source=ds,
                                                                             model_access=self.geppetto_model_access)

                    query_string = query.countQuery if self.count else query.query

                    processed_query_string = template.process_template(dss.get_template(),
                                                                       ID=self.node_id if self.node_id else '',
                                                                       QUERY=query_string,
                                                                       **self.processing_output_map)

                    method = dss.get_connection_type()

                    response = stream_requests(url=ds.url, params=json.loads(processed_query_string), method=method)

                    self.results = dss.process_response(response=response)

            except GeppettoDataSourceException as e:
                raise GeppettoVisitingException(f"Data source exception while running query {query.id}") from e
            except GeppettoInitializationException as e:
                raise GeppettoVisitingException(f"Initialization exception while running query {query.id}") from e
            except Exception as e:
                raise GeppettoVisitingException(f"Unexpected error while running query {query.id}") from e

    def merge_results(self, processed_results: QueryResults):
        #  if this arrives from a first query results should be empty, so we automatically assign
        #  processedResults to results
        if self.results != None:
            if not ID in self.results.header or not ID in processed_results.header:
                raise GeppettoDataSourceException("Cannot merge without an ID in the results")

            id_pos = self.results.header.index(ID)
            proc_id_pos = processed_results.header.index(ID)

            current_record_ids = [record.values[id_pos] for record in self.results.results]
            self.results.header.update(processed_results.header)
            for record in processed_results.results:
                if not record.values[proc_id_pos] in current_record_ids:
                    self.results.results.add(record)
                else:
                    index = current_record_ids.index(record.values[proc_id_pos])

                    values = set(record.values)
                    values.update(self.results.results[index].values)
                    self.results.results[index].values.clear()
                    self.results.results[index].values.extend(values)
        else:
            self.results = processed_results

    def get_datasource(self, query: Query):
        parent = query.eContainer()
        while not isinstance(parent, DataSource):
            parent = parent.eContainer()
        return parent



#!/usr/bin/env python
""" generated source for module WebsocketConnection """
from __future__ import print_function

import json
import logging

from pyecore.valuecontainer import EList
from pygeppetto.api.messaging import TransportMessageFactory
from pygeppetto.constants import UserPrivileges, GeppettoErrorCodes
from pygeppetto.managers import GeppettoManager
from pygeppetto.model.exceptions import GeppettoExecutionException, GeppettoInitializationException
from pygeppetto.model.model_serializer import GeppettoSerializer
from pygeppetto.services.data_manager import DataManagerHelper

from . import inbound_messages as InboundMessages
from . import outbound_messages as OutboundMessages


class GeppettoHandlerTypedException(Exception):
    def __init__(self, msg_type=OutboundMessages.ERROR, msg='Error not specified', exc=None):
        Exception.__init__(self, str(msg))
        self.payload = msg.__dict__ if hasattr(msg, '__dict__') else {"error": str(msg)}
        self.exc = exc
        self.msg_type = msg_type


class Error(object):
    """ generated source for class Error """

    def __init__(self, errorCode, errorMessage, jsonExceptionMsg, id):
        """ generated source for method __init__ """
        self.error_code = errorCode.__str__()
        self.message = json.dumps(errorMessage)
        self.exception = jsonExceptionMsg
        self.id = id


in_out_msg_lookup = {
    InboundMessages.GEPPETTO_VERSION: OutboundMessages.GEPPETTO_VERSION,
    InboundMessages.RESOLVE_IMPORT_TYPE: OutboundMessages.IMPORT_TYPE_RESOLVED,
    InboundMessages.RESOLVE_IMPORT_VALUE: OutboundMessages.IMPORT_VALUE_RESOLVED,
    InboundMessages.RUN_QUERY: OutboundMessages.RETURN_QUERY,
    InboundMessages.RUN_QUERY_COUNT: OutboundMessages.RETURN_QUERY_COUNT,
    InboundMessages.FETCH_VARIABLE: OutboundMessages.VARIABLE_FETCHED,
    InboundMessages.FETCH: OutboundMessages.FETCHED
}


def lookup_return_msg_type(msg_type):
    if msg_type in in_out_msg_lookup:
        return in_out_msg_lookup[msg_type]
    raise RuntimeError("{} not defined into in_out_msg_lookup".format(msg_type))


class GeppettoMessageHandler:
    '''
    Generic message handler: can be subclassed to implement with websocket or other bidirectional means.
    '''
    connection_id = 0
    CLIENT_ID = 'Connection1'

    PRIVILEGES = json.dumps({
        "userName": "Python User",
        "loggedIn": True,
        "hasPersistence": False,
        "privileges": list(up.value for up in UserPrivileges)
    })

    simulationServerConfig = None

    def __init__(self):
        self.geppettoManager = GeppettoManager()
        self.geppettoProject = None

    def sendPrivileges(self):
        self.send_message(None, OutboundMessages.USER_PRIVILEGES, self.PRIVILEGES)

    def sendClientId(self):
        self.send_message(None, OutboundMessages.CLIENT_ID, self.CLIENT_ID)

    def send_message_data(self, msg_data):
        raise NotImplemented

    def handle_message(self, payload):
        '''
        Handles a message coming from the client
        :param payload: a dictionary
        :return:
        '''

        assert 'type' in payload, 'Websocket without type received: {}'.format(payload)

        logging.info('Websocket message received: {}'.format(payload['type']))

        experimentId = -1
        #  de-serialize JSON
        gmsg = payload
        requestID = gmsg['requestID']
        msg_data = None
        #  switch on messages type
        #  NOTE: each messages handler knows how to interpret the GeppettoMessage data field
        msg_type = gmsg['type'].lower()
        try:
            if msg_type == InboundMessages.GEPPETTO_VERSION:
                msg_data = self.getVersionNumber(requestID)

            elif msg_type == InboundMessages.RESOLVE_IMPORT_VALUE:
                received_object = json.loads(gmsg['data'])
                msg_data = self.resolveImportValue(requestID, received_object['projectId'],
                                                   received_object['experimentId'],
                                                   received_object['path'])

            elif msg_type == InboundMessages.RESOLVE_IMPORT_TYPE:
                received_object = gmsg['data']
                msg_data = self.resolveImportType(requestID, received_object.projectId,
                                                  received_object.paths)
            elif msg_type == InboundMessages.LOAD_PROJECT_FROM_CONTENT:
                self.loadProjectFromContent(requestID, gmsg['data'])
            elif msg_type == InboundMessages.LOAD_PROJECT_FROM_URL:
                self.loadProjectFromUrl(requestID, gmsg['data'])
            elif msg_type == InboundMessages.RUN_QUERY:
                received_object = json.loads(gmsg['data'])
                msg_data = self.run_query(received_object)
            elif msg_type == InboundMessages.FETCH_VARIABLE:
                received_object = json.loads(gmsg['data'])
                msg_data = self.fetch_variable(received_object)
            elif msg_type == InboundMessages.FETCH:
                received_object = json.loads(gmsg['data'])
                msg_data = self.fetch(received_object)

            # TODO Complete API. From here on, implementation is not complete: just bare porting from Java
            elif msg_type == InboundMessages.USER_PRIVILEGES:
                msg_data = self.checkUserPrivileges(requestID)
            elif msg_type == InboundMessages.NEW_EXPERIMENT:
                parameters = gmsg['data']
                projectId = int(parameters.get("projectId"))
                msg_data = self.newExperiment(requestID, projectId)
            elif msg_type == InboundMessages.NEW_EXPERIMENT_BATCH:
                received_object = gmsg['data']
                msg_data = self.newExperimentBatch(requestID, received_object.projectId, received_object)
            elif msg_type == InboundMessages.CLONE_EXPERIMENT:
                parameters = gmsg['data']
                projectId = int(parameters.get("projectId"))
                experimentId = int(parameters.get("experimentId"))
                msg_data = self.cloneExperiment(requestID, projectId, experimentId)

            elif msg_type == InboundMessages.LOAD_PROJECT_FROM_ID:
                parameters = gmsg['data']
                if parameters.containsKey("experimentId"):
                    experimentId = int(parameters.get("experimentId"))
                projectId = int(parameters.get("projectId"))
                self.loadProjectFromId(requestID, projectId, experimentId)

            elif msg_type == InboundMessages.PERSIST_PROJECT:
                parameters = gmsg['data']
                projectId = int(parameters.get("projectId"))
                msg_data = self.persistProject(requestID, projectId)
            elif msg_type == InboundMessages.MAKE_PROJECT_PUBLIC:
                parameters = gmsg['data']
                projectId = int(parameters.get("projectId"))
                isPublic = bool(parameters.get("isPublic"))
                msg_data = self.makeProjectPublic(requestID, projectId, isPublic)
            elif msg_type == InboundMessages.DOWNLOAD_PROJECT:
                parameters = gmsg['data']
                projectId = int(parameters.get("projectId"))
                msg_data = self.downloadProject(requestID, projectId)
            elif msg_type == InboundMessages.SAVE_PROJECT_PROPERTIES:
                received_object = gmsg['data']
                msg_data = self.saveProjectProperties(requestID, received_object.projectId,
                                                      received_object.properties)
            elif msg_type == InboundMessages.SAVE_EXPERIMENT_PROPERTIES:
                received_object = gmsg['data']
                msg_data = self.saveExperimentProperties(requestID, received_object.projectId,
                                                         received_object.experimentId,
                                                         received_object.properties)
            elif msg_type == InboundMessages.LOAD_EXPERIMENT:
                parameters = gmsg['data']
                experimentId = int(parameters.get("experimentId"))
                projectId = int(parameters.get("projectId"))
                msg_data = self.loadExperiment(requestID, experimentId, projectId)
            elif msg_type == InboundMessages.GET_SCRIPT:
                received_object = gmsg['data']
                msg_data = self.sendScriptData(requestID, received_object.projectId,
                                               received_object.scriptURL,
                                               self.websocketConnection)
            elif msg_type == InboundMessages.GET_DATA_SOURCE_RESULTS:
                url = None
                dataSourceName = None
                try:
                    parameters = gmsg['data']
                    url = parameters.get("url")
                    dataSourceName = parameters.get("data_source_name")
                    # TODO: use self.websocketConnection when it is implemented
                    msg_data = self.sendDataSourceResults(requestID, dataSourceName, url,
                                                          'self.websocketConnection')
                except IOError as e:
                    self.send_message(requestID, OutboundMessages.ERROR_READING_SCRIPT, "")
            elif msg_type == InboundMessages.GET_EXPERIMENT_STATE:
                received_object = gmsg['data']
                msg_data = self.getExperimentState(requestID, received_object.experimentId,
                                                   received_object.projectId,
                                                   received_object.variables)
            elif msg_type == InboundMessages.DELETE_EXPERIMENT:
                received_object = gmsg['data']
                msg_data = self.deleteExperiment(requestID, received_object.experimentId,
                                                 received_object.projectId)
            elif msg_type == InboundMessages.RUN_EXPERIMENT:
                received_object = gmsg['data']
                msg_data = self.runExperiment(requestID, received_object.experimentId,
                                              received_object.projectId)
            elif msg_type == InboundMessages.SET_WATCHED_VARIABLES:
                received_object = gmsg['data']
                try:
                    msg_data = self.setWatchedVariables(requestID, received_object.variables,
                                                        received_object.experimentId,
                                                        received_object.projectId,
                                                        received_object.watch)
                except GeppettoExecutionException as e:
                    self.send_message(requestID, OutboundMessages.ERROR_SETTING_WATCHED_VARIABLES, "")
                except GeppettoInitializationException as e:
                    self.send_message(requestID, OutboundMessages.ERROR_SETTING_WATCHED_VARIABLES, "")
            elif msg_type == InboundMessages.GET_SUPPORTED_OUTPUTS:
                parameters = gmsg['data']
                experimentId = int(parameters.get("experimentId"))
                projectId = int(parameters.get("projectId"))
                instancePath = parameters.get("instancePath")
                msg_data = self.getSupportedOuputs(requestID, instancePath, experimentId, projectId)
            elif msg_type == InboundMessages.DOWNLOAD_MODEL:
                parameters = gmsg['data']
                experimentId = int(parameters.get("experimentId"))
                projectId = int(parameters.get("projectId"))
                instancePath = parameters.get("instancePath")
                format = parameters.get("format")
                msg_data = self.downloadModel(requestID, instancePath, format, experimentId, projectId)
            elif msg_type == InboundMessages.SET_PARAMETERS:
                received_object = gmsg['data']
                msg_data = self.setParameters(requestID, received_object.modelParameters,
                                              received_object.projectId,
                                              received_object.experimentId)
            elif msg_type == InboundMessages.SET_EXPERIMENT_VIEW:
                received_object = gmsg['data']
                msg_data = self.setExperimentView(requestID, received_object.view,
                                                  received_object.projectId,
                                                  received_object.experimentId)
            elif msg_type == InboundMessages.LINK_DROPBOX:
                parameters = gmsg['data']
                key = parameters.get("key")
                msg_data = self.linkDropBox(requestID, key)
            elif msg_type == InboundMessages.GET_DROPBOX_TOKEN:
                # ReceivedObject received_object = new gmsg['data'], ReceivedObject.class);
                msg_data = self.getDropboxToken(requestID)
            elif msg_type == InboundMessages.UNLINK_DROPBOX:
                parameters = gmsg['data']
                key = parameters.get("key")
                msg_data = self.unLinkDropBox(requestID, key)
            elif msg_type == InboundMessages.UPLOAD_MODEL:
                parameters = gmsg['data']
                experimentId = int(parameters.get("experimentId"))
                projectId = int(parameters.get("projectId"))
                format = parameters.get("format")
                aspectPath = parameters.get("aspectPath")
                msg_data = self.uploadModel(aspectPath, projectId, experimentId, format)
            elif msg_type == InboundMessages.UPLOAD_RESULTS:
                parameters = gmsg['data']
                experimentId = int(parameters.get("experimentId"))
                projectId = int(parameters.get("projectId"))
                format = parameters.get("format")
                aspectPath = parameters.get("aspectPath")
                msg_data = self.uploadResults(aspectPath, projectId, experimentId, format)
            elif msg_type == InboundMessages.DOWNLOAD_RESULTS:
                parameters = gmsg['data']
                experimentId = int(parameters.get("experimentId"))
                projectId = int(parameters.get("projectId"))
                format = parameters.get("format")
                aspectPath = parameters.get("aspectPath")
                msg_data = self.downloadResults(requestID, aspectPath, projectId, experimentId, format)
            elif msg_type == InboundMessages.EXPERIMENT_STATUS:
                msg_data = self.checkExperimentStatus(requestID, gmsg['data'])



            elif msg_type == InboundMessages.RUN_QUERY_COUNT:
                received_object = gmsg['data']
                msg_data = self.runQueryCount(requestID, received_object.projectId,
                                              self.convertRunnableQueriesDataTransferModel(
                                                  received_object.runnableQueries))
            else:
                pass

        except GeppettoHandlerTypedException as e:
            logging.error('Error occurred during message {} handling'.format(msg_type), exc_info=True)
            self.send_message(requestID, e.msg_type, e.payload)

        except Exception as e:
            logging.error('Unexpected error occurred during message {} handling'.format(msg_type), exc_info=True)
            self.send_message(requestID, OutboundMessages.ERROR, {"error": str(e)})

        if msg_data is not None:
            return_msg_type = lookup_return_msg_type(msg_type)
            self.send_message(requestID, return_msg_type, msg_data)

    def run_query(self, received_object):

        geppetto_project = self.retrieveGeppettoProject(received_object['projectId'])
        runnable_queries = self.convertRunnableQueriesDataTransferModel(received_object['runnableQueries'])
        results = self.geppettoManager.run_query(runnable_queries, geppetto_project)
        msg_data = GeppettoSerializer.serialize(results, True)
        return msg_data

    def fetch_variable(self, received_object):
        try:

            geppetto_project = self.retrieveGeppettoProject(received_object['projectId'])
            results = self.geppettoManager.fetch_variable(received_object['dataSourceId'],
                                                          received_object['variableId'],
                                                          geppetto_project)
            return GeppettoSerializer.serialize(results, True)
        except Exception as e:
            self.error(e, "Error fetching variable " + str(received_object['variableId']))

    def fetch(self, received_object):
        try:
            geppetto_project = self.retrieveGeppettoProject(received_object['projectId'])
            results = self.geppettoManager.fetch(received_object['dataSourceId'],
                                                 received_object['variables'] if 'variables' in received_object else [],
                                                 received_object['instances'] if 'instances' in received_object else [],
                                                 geppetto_project)
            return GeppettoSerializer.serialize(results, True)
        except Exception as e:
            self.error(e, "Error fetching" + str(received_object))

    def send_message(self, requestID, return_msg_type, msg_data):
        msg_data = TransportMessageFactory.getTransportMessage(requestID=requestID, type_=return_msg_type,
                                                               update=msg_data).__dict__
        self.send_message_data(msg_data)

    def loadProjectFromUrl(self, requestID, urlString):
        data_manager = DataManagerHelper.getDataManager()
        try:
            geppettoProject = data_manager.get_project_from_url(urlString)
        except Exception as e:
            raise GeppettoHandlerTypedException(OutboundMessages.ERROR_LOADING_PROJECT,
                                                f"Error while loading project from url: {urlString}: {e}")
        if geppettoProject == None:
            raise GeppettoHandlerTypedException(OutboundMessages.ERROR_LOADING_PROJECT,
                                                "Project not found")
        else:
            return self.loadGeppettoProject(requestID, geppettoProject, -1)

    def loadProjectFromContent(self, requestID, projectContentJSON):
        dataManager = DataManagerHelper.getDataManager()

        geppettoProject = dataManager.getProjectFromJson(projectContentJSON)
        self.loadGeppettoProject(requestID, geppettoProject, -1)

    def loadGeppettoProject(self, requestID, geppettoProject, experimentId):

        try:
            readOnly = geppettoProject.volatile  # TODO implement logic related to user projects which are not readonly

            self.geppettoManager.load_project(geppettoProject)
            # Here the project is loaded: a runtime project is created with its model

            geppettoProject.geppettoModel = {
                'id': geppettoProject.geppettoModel.id,
                'type': 'MODEL'
            }  # There is something odd here: we are  sending the project without the model, although the model is there. It's the same in Java anyway, we are just having a PersistedData placeholder there. Why not serialize the model together with the project?

            project_message_update = json.dumps({
                'persisted': not geppettoProject.volatile,
                'project': geppettoProject.__dict__,
                'isReadOnly': readOnly
            })
            self.send_message(requestID, OutboundMessages.PROJECT_LOADED, project_message_update)

            runtime_project = self.geppettoManager.get_runtime_project(geppettoProject)

            geppettoModelJSON = GeppettoSerializer.serialize(
                runtime_project.model, False)
            self.send_message(requestID, OutboundMessages.GEPPETTO_MODEL_LOADED, geppettoModelJSON)

            GeppettoSerializer.serialize(
                runtime_project.model, True)  # This is setting synched to true on the model objects

            # TODO handle experiment
        except Exception as e:
            self.error(e, "Could not load geppetto project")

    def loadProjectFromId(self, requestID, projectId, experimentId):
        """ generated source for method loadProjectFromId """
        dataManager = DataManagerHelper.getDataManager()
        try:
            geppettoProject = dataManager.getGeppettoProjectById(projectId)
            if self.geppettoProject == None:
                raise GeppettoHandlerTypedException(OutboundMessages.ERROR_LOADING_PROJECT,
                                                    "Project not found")
            else:
                self.loadGeppettoProject(requestID, self.geppettoProject, experimentId)
        except Exception as e:
            raise GeppettoHandlerTypedException(OutboundMessages.ERROR_LOADING_PROJECT, str(e), e)

    def resolveImportType(self, requestID, projectId, typePaths):
        """ generated source for method resolveImportType """
        geppettoProject = self.retrieveGeppettoProject(projectId)
        try:
            geppettoModel = self.geppettoManager.resolve_import_type(typePaths, geppettoProject)
            return GeppettoSerializer.serializeToJSON(geppettoModel, True)
        except IOError as e:
            self.error(e, "Error importing type " + typePaths)
        except GeppettoExecutionException as e:
            self.error(e, "Error importing type " + typePaths)

    def resolveImportValue(self, requestID, projectId, experimentId, path):
        """ generated source for method resolveImportValue """
        geppettoProject = self.retrieveGeppettoProject(projectId)
        experiment = self.retrieveExperiment(experimentId, geppettoProject)
        try:
            geppettoModel = self.geppettoManager.resolve_import_value(path, geppettoProject, experiment)
            return GeppettoSerializer.serialize(geppettoModel, True)
        except IOError as e:
            self.error(e, "Error importing value " + path)
        except GeppettoExecutionException as e:
            self.error(e, "Error importing value " + path)

    def retrieveExperiment(self, experimentID, geppettoProject):
        """ generated source for method retrieveExperiment """
        theExperiment = None
        #  Look for experiment that matches id passed
        for e in geppettoProject.experiments:
            if e.getId() == experimentID:
                #  The experiment is found
                theExperiment = e
                break
        return theExperiment

    def retrieveGeppettoProject(self, projectId):
        """ generated source for method retrieveGeppettoProject """
        dataManager = DataManagerHelper.getDataManager()
        return dataManager.getGeppettoProjectById(projectId)

    def error(self, exception: Exception, errorMessage):
        """ generated source for method error """
        exceptionMessage = ""
        if exception != None:
            exceptionMessage = str(exception)
        error = Error(GeppettoErrorCodes.EXCEPTION, errorMessage, exceptionMessage, 0)
        logging.error(errorMessage, exception)

        raise GeppettoHandlerTypedException(OutboundMessages.ERROR, errorMessage) from exception

    def info(self, requestID, message):
        """ generated source for method info """
        logging.info(message)
        raise GeppettoHandlerTypedException(OutboundMessages.INFO_MESSAGE, message)

    def setConnectionProject(self, geppettoProject):
        """ generated source for method setConnectionProject """
        if self.geppettoProject != None:
            self.geppettoManager.close_project(self.geppettoProject)
        self.geppettoProject = geppettoProject

    def getVersionNumber(self, requestID):
        import pkg_resources  # part of setuptools
        return pkg_resources.require("pygeppetto")[0].version

    def convertRunnableQueriesDataTransferModel(self, runnableQueries):
        from pygeppetto.model.datasources.datasources import RunnableQuery
        return [RunnableQuery(**dt) for dt in runnableQueries]



"""Definition of meta model 'instances'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *
from ..model import Node, ISynchable
from ..values import Connectivity


name = 'instances'
nsURI = 'https://raw.githubusercontent.com/openworm/org.geppetto.model/master/src/main/resources/geppettoModel.ecore#//instances'
nsPrefix = 'gep'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)


@abstract
class Instance(Node):

    type = EReference(ordered=True, unique=True, containment=False)
    value = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, type=None, value=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if value is not None:
            self.value = value


class SimpleInstance(Instance):

    visualValue = EReference(ordered=True, unique=True, containment=True)
    position = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, visualValue=None, position=None, **kwargs):

        super().__init__(**kwargs)

        if visualValue is not None:
            self.visualValue = visualValue

        if position is not None:
            self.position = position


class SimpleConnectionInstance(Instance):

    connectivity = EAttribute(eType=Connectivity, derived=False, changeable=True)
    a = EReference(ordered=True, unique=True, containment=False)
    b = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, a=None, b=None, connectivity=None, **kwargs):

        super().__init__(**kwargs)

        if connectivity is not None:
            self.connectivity = connectivity

        if a is not None:
            self.a = a

        if b is not None:
            self.b = b



from enum import Enum, unique

from pyecore.ecore import EPackage


@unique
class GeppettoErrorCodes(Enum):
    """ generated source for enum GeppettoErrorCodes """
    INITIALIZATION = u'INITIALIZATION'
    EXCEPTION = u'EXCEPTION'
    MODEL_INTERPRETER = u'MODEL_INTERPRETER'
    SIMULATOR = u'SIMULATOR'
    SIMULATION = u'SIMULATION'
    GENERIC = u'GENERIC'


@unique
class UserPrivileges(Enum):
    READ_PROJECT = 'READ_PROJECT'
    WRITE_PROJECT = 'WRITE_PROJECT'
    RUN_EXPERIMENT = 'RUN_EXPERIMENT'
    DROPBOX_INTEGRATION = 'DROPBOX_INTEGRATION'
    DOWNLOAD = 'DOWNLOAD'
    ADMIN = 'ADMIN'


@unique
class ExperimentStatus(Enum):
    DESIGN = 'DESIGN'
    QUEUED = 'QUEUED'
    RUNNING = 'RUNNING'
    ERROR = 'ERROR'
    COMPLETED = 'COMPLETED'
    DELETED = 'DELETED'
    CANCELED = 'CANCELED'


@unique
class GeppettoFeature(Enum):
    """ generated source for enum GeppettoFeature """
    SET_PARAMETERS_FEATURE = u'SET_PARAMETERS_FEATURE'
    DEFAULT_VIEW_CUSTOMISER_FEATURE = u'DEFAULT_VIEW_CUSTOMISER_FEATURE'


class GeppettoPackage(EPackage):
    # FIXME GeppettoPackage should be generated by eCore?

    eNAME = "model"
    eNS_URI = "https://raw.githubusercontent.com/openworm/org.geppetto.model/development/src/main/resources/geppettoModel.ecore"

    eNS_URI_TEMPLATE = "https://raw.githubusercontent.com/openworm/org.geppetto.model/$VERSION$/src/main/resources/geppettoModel.ecore"

    eNS_PREFIX = "gep"

    eINSTANCE = None  # TODO no instance of GeppettoPackage yet implemented

    GEPPETTO_MODEL = 0

    GEPPETTO_MODEL__VARIABLES = 0

    GEPPETTO_MODEL__LIBRARIES = 1

    GEPPETTO_MODEL__TAGS = 2

    GEPPETTO_MODEL__ID = 3

    GEPPETTO_MODEL__NAME = 4

    GEPPETTO_MODEL__DATA_SOURCES = 5

    GEPPETTO_MODEL__QUERIES = 6

    GEPPETTO_MODEL_FEATURE_COUNT = 7

    GEPPETTO_MODEL_OPERATION_COUNT = 0

    ISYNCHABLE = 11

    ISYNCHABLE__SYNCHED = 0

    ISYNCHABLE_FEATURE_COUNT = 1

    ISYNCHABLE_OPERATION_COUNT = 0

    NODE = 1

    NODE__SYNCHED = ISYNCHABLE__SYNCHED

    NODE__ID = ISYNCHABLE_FEATURE_COUNT + 0

    NODE__NAME = ISYNCHABLE_FEATURE_COUNT + 1

    NODE__TAGS = ISYNCHABLE_FEATURE_COUNT + 2

    NODE_FEATURE_COUNT = ISYNCHABLE_FEATURE_COUNT + 3

    NODE___GET_PATH = ISYNCHABLE_OPERATION_COUNT + 0

    NODE_OPERATION_COUNT = ISYNCHABLE_OPERATION_COUNT + 1

    GEPPETTO_LIBRARY = 2

    GEPPETTO_LIBRARY__SYNCHED = NODE__SYNCHED

    GEPPETTO_LIBRARY__ID = NODE__ID

    GEPPETTO_LIBRARY__NAME = NODE__NAME

    GEPPETTO_LIBRARY__TAGS = NODE__TAGS

    GEPPETTO_LIBRARY__TYPES = NODE_FEATURE_COUNT + 0

    GEPPETTO_LIBRARY__SHARED_TYPES = NODE_FEATURE_COUNT + 1

    GEPPETTO_LIBRARY_FEATURE_COUNT = NODE_FEATURE_COUNT + 2

    GEPPETTO_LIBRARY___GET_PATH = NODE___GET_PATH

    GEPPETTO_LIBRARY___GET_TYPE_BY_ID = NODE_OPERATION_COUNT + 0

    GEPPETTO_LIBRARY_OPERATION_COUNT = NODE_OPERATION_COUNT + 1

    LIBRARY_MANAGER = 3

    LIBRARY_MANAGER__LIBRARIES = 0

    LIBRARY_MANAGER_FEATURE_COUNT = 1



from pyecore.resources import global_registry
from .instances import getEClassifier, eClassifiers
from .instances import name, nsURI, nsPrefix, eClass
from .instances import Instance, SimpleInstance, SimpleConnectionInstance

from ..values import VisualValue, Point, Value
from ..types import Type
from ..model import Tag

from . import instances
from .. import model


__all__ = ['Instance', 'SimpleInstance', 'SimpleConnectionInstance']

eSubpackages = []
eSuperPackage = model
instances.eSubpackages = eSubpackages
instances.eSuperPackage = eSuperPackage


otherClassifiers = []

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for classif in eClassifiers.values():
    eClass.eClassifiers.append(classif.eClass)

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)

register_packages = [instances] + eSubpackages
for pack in register_packages:
    global_registry[pack.nsURI] = pack



import json
import os.path

from deprecated import deprecated

from pyecore.resources import ResourceSet, URI
from pygeppetto.model import MDTimeSeries, Value, VisualValue, AArrayValue, SimpleInstance, SimpleConnectionInstance
from pygeppetto.model.values import Connectivity
from pygeppetto.utils import clone

from .model import GeppettoModel
from .values import Cylinder, Sphere, Point, PhysicalQuantity, TimeSeries, Unit, ImportValue, Text, JSON, \
    ArrayValue, HTML, URL, Expression, Dynamics, Image, Connection, Particles, Pointer, Metadata
from .variables import Variable, TypeToValueMap


class GeppettoCommonLibrary:
    TYPE_PARAMETER = 0
    TYPE_DYNAMICS = 1
    TYPE_STATE_VARIABLE = 2
    TYPE_HTML = 3
    TYPE_URL = 4
    TYPE_TEXT = 5
    TYPE_POINT = 6
    TYPE_EXPRESSION = 7
    TYPE_VISUAL = 8
    TYPE_POINTER = 9
    TYPE_IMAGE = 10
    TYPE_CONNECTION = 11
    TYPE_PARTICLES = 12
    TYPE_JSON = 13
    TYPE_SIMPLE_ARRAY = 14
    TYPE_METADATA = 15
    TYPE_EDGE = 16
    TYPE_NODE = 17

    rset = ResourceSet()
    # Build the model URI
    model_uri = URI(os.path.join(os.path.dirname(__file__), '..',
                                 'ecore', 'GeppettoCommonLibrary.xmi'))
    resource = rset.get_resource(model_uri)  # We load the model
    instance = resource.contents[0]

    @classmethod
    def instance_copy(cls):
        return clone(cls.instance)



class SharedLibraryManager:

    @staticmethod
    def get_shared_common_library():
        return GeppettoCommonLibrary.instance_copy()

class GeppettoModelFactory:


    def __init__(self, geppetto_common_library=None):
        if geppetto_common_library is None:
            geppetto_common_library = SharedLibraryManager.get_shared_common_library()
        self.geppetto_common_library = geppetto_common_library

    @classmethod
    def createGeppettoModel(cls, name):
        return GeppettoModel(name=name, libraries=(SharedLibraryManager.get_shared_common_library(),))

    def create_variable(self, id, cl_type, raw_initial_value):
        variable = Variable(id=id, name=id)
        _type = self.geppetto_common_library.types[cl_type] if isinstance(cl_type, int) else cl_type
        variable.types.append(self.geppetto_common_library.types[cl_type])
        variable.initialValues.append(TypeToValueMap(_type, raw_initial_value))
        return variable

    def create_simple_instance(self, id, cl_type, value=None):
        _type = self.geppetto_common_library.types[cl_type] if isinstance(cl_type, int) else cl_type
        return SimpleInstance(id=id, name=id, type=_type, value=value)

    def create_connection_instance(self, id, cl_type, a, b, connectivity: Connectivity, value=None):
        _type = self.geppetto_common_library.types[cl_type] if isinstance(cl_type, int) else cl_type
        return SimpleConnectionInstance(id=id, name=id, type=_type, a=a, b=b, connectivity=connectivity, value=value)

    @deprecated(version='0.6', reason='Use the snake case version')
    def createCylinder(self, id, bottomRadius=1.0, topRadius=1.0,
                       position=None, distal=None):

        return self.create_cylinder(id, bottomRadius, topRadius,
                                    position, distal)

    def create_cylinder(self, id, bottom_radius=1.0, top_radius=1.0,
                        position=None, distal=None):
        position = position or Point()
        distal = distal or Point()
        cylinder = Cylinder(bottomRadius=bottom_radius, topRadius=top_radius)
        cylinder.distal = distal
        cylinder.position = position

        return self.create_variable(id, GeppettoCommonLibrary.TYPE_VISUAL, cylinder)

    @deprecated(version='0.6', reason='Use the snake case version')
    def createSphere(self, id, radius=1.0, position=None):
        return self.create_sphere(id, radius, position)

    def create_sphere(self, id, radius=1.0, position=None):
        position = position or Point()
        variable = Variable(id=id)
        variable.types.append(self.geppetto_common_library.types[GeppettoCommonLibrary.TYPE_VISUAL])
        sphere = Sphere(radius=radius, position=position)
        variable.initialValues.append(
            TypeToValueMap(self.geppetto_common_library.types[GeppettoCommonLibrary.TYPE_VISUAL], sphere))
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_VISUAL, sphere)

    @classmethod
    @deprecated(version='0.6', reason='Use the snake case version')
    def createTimeSeries(cls, id=None, values=None, unit=None):
        return cls.create_time_series(values, unit)


    @classmethod
    def create_time_series(cls, values, unit=None):
        if unit:
            unit = Unit(unit)
        ts = TimeSeries(value=values, unit=unit)
        return ts

    @classmethod
    @deprecated(version='0.6', reason='no need to use the factory for this')
    def createMDTimeSeries(self, id, values):
        return MDTimeSeries(value=values)

    @deprecated(version='0.6', reason='no need to use the factory for this')
    def createImportValue(self):
        iv = ImportValue()
        return iv

    def create_time_series_variable(self, id, values, unit=None):
        if unit:
            unit = Unit(unit)
        ts = TimeSeries(value=values, unit=unit)
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_STATE_VARIABLE, ts)

    @deprecated(version='0.6', reason='Use the snake case version')
    def createMDTimeSeriesVariable(self, id, values):
        MD_ts = MDTimeSeries(value=values)
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_STATE_VARIABLE, MD_ts)

    def create_MD_time_series_variable(self, id, values):
        MD_ts = MDTimeSeries(value=values)
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_STATE_VARIABLE, MD_ts)

    @deprecated(version='0.6', reason='Use the snake case version')
    def createStateVariable(self, id, initialValue=None):
        return self.create_state_variable(id, initialValue)

    def create_state_variable(self, id, initialValue=None):
        initialValue = initialValue or PhysicalQuantity()
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_STATE_VARIABLE, initialValue)

    def create_parameter_variable(self, id, initialValue=None):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_PARAMETER, initialValue)

    def create_dynamics_variable(self, id, dynamics=None, initialCondition=None):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_DYNAMICS,
                                    Dynamics(dynamics=dynamics, initialCondition=initialCondition))

    def create_html_variable(self, id, html_text):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_HTML, HTML(html_text))

    def create_url_variable(self, id, url):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_URL, URL(url))

    def create_text_variable(self, id, text=''):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_TEXT, Text(text))

    def create_point_variable(self, id, x=1.0, y=1.0, z=1.0):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_POINT, Point(x, y, z))

    def create_expression_variable(self, id, expression):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_EXPRESSION, Expression(expression))

    def create_visual_variable(self, id, initialValue: VisualValue = None):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_VISUAL, initialValue)

    def create_image_variable(self, id, data=None, name=None, reference=None, format=None, initialValue=None):
        if data or name or reference or format:
            initialValue = Image(data=data, name=name, reference=reference, format=format)
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_IMAGE, initialValue)

    def create_connection_variable(self, id, initialValue: Connection = None):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_CONNECTION, initialValue)

    def create_particles_variable(self, id, particles):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_PARTICLES, Particles(particles=particles))

    def create_pointer_variable(self, id, initialValue: Pointer):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_POINTER, initialValue)

    def create_JSON_variable(self, id, serializable_obj):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_JSON, JSON(json.dumps(serializable_obj)))

    def create_simple_array_variable(self, id, initialValue: AArrayValue):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_SIMPLE_ARRAY, initialValue)

    def create_metadata_variable(self, id, initialValue: Metadata):
        return self.create_variable(id, GeppettoCommonLibrary.TYPE_METADATA, initialValue)

    def create_node_instance(self, id, value):
        return self.create_simple_instance(id, GeppettoCommonLibrary.TYPE_NODE, value)

    def create_edge_instance(self, id, a, b, connectivity: Connection, value=None):
        return self.create_connection_instance(id, GeppettoCommonLibrary.TYPE_EDGE, a=a, b=b, connectivity=connectivity, value=value)



class GeppettoProject(object):

    def __init__(self, id, name, geppetto_model, volatile=True, base_url=None, public=False, experiments=None,
                 view=None):
        self.id = id
        self.name = name
        self.geppettoModel = geppetto_model  # Beware must use the camelCase here otherwise we cannot import from JSON
        self.base_url = base_url
        self.experiments = list(experiments) if experiments is not None else []
        self.active_experiment_id = -1 if not experiments else 0
        self.volatile = volatile
        self.view = view
        self.is_public = public

    def __eq__(self, o: object) -> bool:
        return self.id == o.id if hasattr(o, 'id') else False

    def __hash__(self) -> int:
        return hash(self.id)

    def __str__(self) -> str:
        return "GeppettoProject[id={}, name={}]".format(self.id, self.name)





from pyecore.resources import global_registry
from .model import nsURI, eSubpackages

from .model import overrides
# Manually register all the URI and master URI in the global registry
# This registering is performed outside the previous 'for' to ease future
# code merging (in case of new metamodel versions)
geppetto_master_uri = ('https://raw.githubusercontent.com/openworm/'
                       'org.geppetto.model/master/src/main/resources/'
                       'geppettoModel.ecore')

global_registry[nsURI] = model
global_registry[geppetto_master_uri] = model
for subpack in eSubpackages:
    global_registry[subpack.nsURI] = subpack
    global_registry[geppetto_master_uri + '#//' + subpack.name] = subpack



from pygeppetto.model import Variable, SimpleInstance
from pygeppetto.model.datasources import DataSource, QueryResults, ProcessQuery, RunnableQuery, BooleanOperator
from pygeppetto.model.exceptions import GeppettoInitializationException
from pygeppetto.model.model_access import GeppettoModelAccess
from pygeppetto.model.utils.datasource import query_check, set_custom_query_result_hash, unset_custom_query_result_hash
from pygeppetto.visitors.data_source_visitors import ExecuteQueryVisitor

ID = "ID"


class GeppettoDataSourceException(Exception): pass


class ServiceCreator(type):
    """
    Collects all subclasses of DataSourceService and enables discovery
    """
    data_source_services = {}

    def __init__(cls, name, bases, dct):
        """Initializing a new DataSourceService class"""
        super().__init__(name, bases, dct)
        if cls.__name__ != 'DataSourceService':
            ServiceCreator.data_source_services[cls.__name__] = cls

    @classmethod
    def get_new_datasource_service_instance(mcs, data_source: DataSource, model_access):
        data_source_discovery_id = data_source.dataSourceService

        return mcs.get_new_service_instance(data_source_discovery_id, data_source, model_access)

    @classmethod
    def get_new_service_instance(mcs, class_name, *params):
        """
        Finds ans instantiates a service registered with this metaclass
        :param class_name: the class name
        :param params:
        :return:
        """
        if not class_name in ServiceCreator.data_source_services:
            raise GeppettoInitializationException(f"The service {class_name} was not found!")
        return mcs.data_source_services[class_name](*params)


class QueryProcessor(metaclass=ServiceCreator):

    def process(self, query: ProcessQuery, data_source: DataSource, variable, results: QueryResults,
                model_access: GeppettoModelAccess) -> QueryResults:
        raise NotImplemented

    def get_processing_output_map(self):
        return {}


class DataSourceService(metaclass=ServiceCreator):

    def __init__(self, configuration: DataSource, model_access: GeppettoModelAccess):
        self.configuration = configuration
        self.model_access = model_access

    def fetch_variable(self, variable_id):
        var_query = self.configuration.fetchVariableQuery
        execute_query_visitor = ExecuteQueryVisitor(node_path=variable_id, geppetto_model_access=self.model_access)
        execute_query_visitor.do_switch(var_query)

    def fetch_instance(self, instance_id):
        var_query = self.configuration.fetchVariableQuery
        execute_query_visitor = ExecuteQueryVisitor(self.model_access, node_path=instance_id)
        execute_query_visitor.do_switch(var_query)

    def execute(self, queries, count_only=False):
        return self.get_results(
            {self.execute_runnable_query(runnable_query): runnable_query.booleanOperator for runnable_query in queries}
        )

    def get_number_of_results(self, queries):
        return len(self.execute(queries, count_only=True).results)

    def get_available_queries(self, variable):
        return [query for query in self.configuration.queries if query_check(query, variable)]

    def execute_runnable_query(self, runnable_query: RunnableQuery, count_only=False):
        """
        Moved from https://github.com/openworm/org.geppetto.datasources/blob/master/src/main/java/org/geppetto/datasources/ExecuteMultipleQueriesVisitor.java
        Implementation is simplified without caching
        :param runnable_query:
        :return:
        """
        query = self.model_access.get_query(runnable_query.queryPath)
        execute_query_visitor = ExecuteQueryVisitor(node_path=runnable_query.targetVariablePath, geppetto_model_access=self.model_access, count_only=count_only)
        execute_query_visitor.do_switch(query)
        return execute_query_visitor.results

    def get_results(self, results: dict) -> QueryResults:
        """
            Ported from https://github.com/openworm/org.geppetto.datasources/blob/master/src/main/java/org/geppetto/datasources/ExecuteMultipleQueriesVisitor.java#getResults
        """
        if not results:
            return QueryResults()
        final_results = QueryResults(header=next(iter(results.keys())).header)
        first = True

        id_index = final_results.header.index(ID)
        set_custom_query_result_hash(id_index)

        for result, operator in results.items():
            if final_results.header != result.header:  # TODO test it: may not be supported
                raise GeppettoDataSourceException(
                    "Multiple queries were executed but they returned incompatible headers"
                )

            if first or operator == BooleanOperator.OR:
                method = final_results.results.update

            elif operator == BooleanOperator.AND:
                method = final_results.results.intersection_update

            elif operator == BooleanOperator.NAND:
                method = final_results.results.difference_update
            else:
                raise GeppettoDataSourceException(f"Missing operator for query result {result}")

            method(r for r in result.results)

            first = False

        unset_custom_query_result_hash()

        return final_results

    def get_connection_type(self):
        raise NotImplementedError("Method get_connection_type must be defined in every DataSourceService"
                                  "and return either `GET` or `POST`.")

    def get_template(self):
        raise NotImplemented

    def process_response(self, response) -> QueryResults:
        """
        Custom logic to process response
        :param response_dict:
        :return:
        """
        raise NotImplemented



import pytest
from unittest.mock import Mock
import json

import responses

from pyecore.resources import ResourceSet, URI
from pygeppetto.api.inbound_messages import InboundMessages
from pygeppetto.api.message_handler import GeppettoMessageHandler
from pygeppetto.data_model import GeppettoProject
from pygeppetto.model import GeppettoModel, QueryResults
from pygeppetto.model.datasources import QueryResult
from pygeppetto.services.data_manager import GeppettoDataManager, DataManagerHelper
from pygeppetto.services.data_source_service import DataSourceService, ServiceCreator

from .test_xmi import filepath
from .mocks import neo4j_response, MockFetchQueryProcessor, MockDataManager, MockDataSourceService





@pytest.fixture
def message_handler():
    return GeppettoMessageHandler()





DataManagerHelper.setDataManager(MockDataManager())


def load_project(message_handler):
    messages = []

    def mock_send_message(message):
        print(message)
        messages.append(message)
        assert not 'error' in message['type']

    # 1 preparation: first we need a RuntimeProject active and model loaded
    message_handler.send_message_data = Mock(side_effect=mock_send_message)
    message_handler.handle_message(
        dict(type=InboundMessages.LOAD_PROJECT_FROM_URL, requestID='whatever', data='this will be ignored'))
    assert message_handler.send_message_data.call_count == 2
    geppetto_manager = message_handler.geppettoManager
    project = message_handler.retrieveGeppettoProject('mock')
    assert project.id == 'mock'
    runtime_project = geppetto_manager.get_runtime_project(project)
    assert len(runtime_project.model.dataSources) == 1
    configured_data_source_service = runtime_project.model.dataSources[0].dataSourceService
    assert configured_data_source_service == MockDataSourceService.__name__
    assert ServiceCreator.get_new_datasource_service_instance(runtime_project.model.dataSources[0],
                                                              None).__class__==MockDataSourceService
    return runtime_project, messages


@responses.activate
def test_run_query(message_handler):
    runtime_project, messages = load_project(message_handler)

    URL = "http://my-neo4j/db/data/transaction"

    responses.add(responses.POST, URL, json=neo4j_response(), status=200)
    msg_data = json.dumps({
        "projectId": 'mock',
        "runnableQueries": [
            {
                "queryPath": "mockDataSource.mock_query",
                "targetVariablePath": "v"
            }
        ]
    })
    run_query_msg = {"requestID": "Connection23-5", "type": InboundMessages.RUN_QUERY,
                     "data": msg_data}
    message_handler.handle_message(run_query_msg)

    assert message_handler.send_message_data.call_count == 3

    result = json.loads(messages[2]['data'])
    model = json.loads(result['return_query'])
    assert model['eClass'] == QueryResults.__name__


@responses.activate
def test_fetch_variable(message_handler):
    runtime_project, messages = load_project(message_handler)

    URL = "http://my-neo4j/db/data/transaction"

    responses.add(responses.POST, URL, json=neo4j_response(), status=200)
    msg_data = json.dumps({
        "projectId": 'mock',
        "dataSourceId": "mockDataSource",
        "variableId": ["myvar"],
    })

    run_query_msg = {"requestID": "Connection23-5", "type": InboundMessages.FETCH_VARIABLE,
                     "data": msg_data}
    message_handler.handle_message(run_query_msg)

    assert message_handler.send_message_data.call_count == 3

    variable = next(var for var in runtime_project.model.worlds[0].variables if var.id == 'myvar')

    assert variable.name == MockFetchQueryProcessor.variable_name

    variable_container = variable.eContainer()
    assert variable_container.id
    assert variable.id
    assert "myvar" in messages[2]['data']


@responses.activate
def test_fetch(message_handler):
    runtime_project, messages = load_project(message_handler)

    URL = "http://my-neo4j/db/data/transaction"

    responses.add(responses.POST, URL, json=neo4j_response(), status=200)
    msg_data = json.dumps({
        "projectId": 'mock',
        "dataSourceId": "mockDataSource",
        "variables": ["myvar"],
        "instances": ["myinst"],
        "worldId": "w"
    })

    run_query_msg = {"requestID": "Connection23-5", "type": InboundMessages.FETCH,
                     "data": msg_data}
    message_handler.handle_message(run_query_msg)

    assert message_handler.send_message_data.call_count == 3

    variable = next(var for var in runtime_project.model.worlds[0].variables if var.id == 'myvar')

    assert variable.name == MockFetchQueryProcessor.variable_name

    variable_container = variable.eContainer()
    assert variable_container.id
    assert variable.id
    assert "myvar" in messages[2]['data']

    instance = next(inst for inst in runtime_project.model.worlds[0].instances if inst.id == 'myinst')
    assert instance.name == MockFetchQueryProcessor.variable_name
    assert "myinst" in messages[2]['data']



from pyecore.resources import global_registry
from .variables import getEClassifier, eClassifiers
from .variables import name, nsURI, nsPrefix, eClass
from .variables import Variable, TypeToValueMap

from ..values import Point, Value
from ..types import Type
from ..model import Tag

from . import variables
from .. import model


__all__ = ['Variable', 'TypeToValueMap']

eSubpackages = []
eSuperPackage = model
variables.eSubpackages = eSubpackages
variables.eSuperPackage = eSuperPackage


otherClassifiers = []

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for classif in eClassifiers.values():
    eClass.eClassifiers.append(classif.eClass)

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)

register_packages = [variables] + eSubpackages
for pack in register_packages:
    global_registry[pack.nsURI] = pack



from pyecore.ecore import EObject
from pygeppetto.visitors import Switch


def apply_single(eobject: EObject, fn, condition=lambda eobject: True):
    for element in eobject.eAllContents():
        if condition(element):
            fn(element)


def apply(eobject: EObject, visitor: Switch):
    visitor.do_switch(eobject)
    for element in eobject.eAllContents():
        visitor.do_switch(element)


def apply_direct_children_only(eobject: EObject, visitor: Switch):
    for element in eobject.eContents:
        visitor.do_switch(element)



import pytest
import pygeppetto.model.utils.pointer_utility as pointer_utility
from pygeppetto.constants import UserPrivileges
from pygeppetto.local_data_model import LocalGeppettoProject, LocalUser, LocalUserGroup
from pygeppetto.managers.geppetto_manager import GeppettoManager
from pygeppetto.model import GeppettoLibrary, CompositeType, StateVariableType
from pygeppetto.model.exceptions import GeppettoModelException
from pygeppetto.model.model_access import GeppettoModelAccess
from pygeppetto.model.types import ImportType
from pygeppetto.model.values import ImportValue
from pygeppetto.services.model_interpreter import add_model_interpreter

from .mocks import MockModelInterpreter
from pygeppetto.model.model_serializer import GeppettoModelSerializer


@pytest.fixture
def model_access():
    model_interpreter = MockModelInterpreter()
    model_library = GeppettoLibrary(name='mocklibrary', id='mocklibrary')
    geppetto_model = model_interpreter.create_model(library=model_library)
    add_model_interpreter(model_library.id, model_interpreter)
    return GeppettoModelAccess(geppetto_model)


def test_access(model_access):
    assert CompositeType == type(model_access.get_type('mocklibrary.ct1'))
    assert StateVariableType == type(model_access.get_type('common.StateVariable'))
    assert model_access.get_variable('v3').id == 'v3'
    assert model_access.get_variable('v3.v31').id == 'v31'
    assert model_access.get_variable('v3.v32').id == 'v32'
    with pytest.raises(GeppettoModelException):
        assert model_access.get_value('v3') is None
    assert ImportValue == type(model_access.get_value('v3.v31'))


def test_swap_type(model_access):
    model = model_access.geppetto_model
    GeppettoModelSerializer.serialize(model, True)
    assert model_access.get_variable('v1').synched == True
    assert model_access.get_variable('v3.v31').synched == True

    # Change the state variable to a composite
    model_access.swap_type(model_access.get_type('common.StateVariable'), model_access.get_type('mocklibrary.ct1'))

    assert model_access.get_variable('v1.v32').id == 'v32'
    assert model_access.get_variable('v1').synched == False
    assert model_access.get_variable('v1.v31').synched == False
    assert model_access.get_variable('v3').synched == True
    model_access.get_type('mocklibrary.ct1').synched = False



import pytest
from pygeppetto.model import GeppettoModel, GeppettoLibrary, Tag


def test_create_pygeppetto_model():
    root = GeppettoModel()
    assert root.name is None


def test_create_pygeppetto_model_name():
    root = GeppettoModel(name='model1')
    assert root.name == 'model1'


def test_create_pygeppetto_readme_test():
    flib = GeppettoLibrary(name='mylib')
    root = GeppettoModel(name='MyGeppettoModel', libraries=[flib])
    assert flib in root.libraries
    assert flib.name == 'mylib'
    assert root.name == 'MyGeppettoModel'


def test_create_pygeppetto_model_tag():
    tag = Tag()
    model = GeppettoModel()
    model.tags.append(tag)
    assert tag in model.tags



""" Derived from generated source for class PointerUtility
https://github.com/openworm/org.geppetto.model/blob/master/src/main/java/org/geppetto/model/util/java"""
import sys

from multimethod import multidispatch

from ..exceptions import GeppettoModelException
from ..model import GeppettoLibrary, GeppettoModel
from ..types import CompositeType, ArrayType, Type
from ..values.values import Pointer, PointerElement
from ..variables import Variable

# Fake a static class
PointerUtility = sys.modules[__name__]


@multidispatch
def get_pointer(variable: Variable, type_, index):
    pointerElement = PointerElement()
    pointerElement.index = index
    pointerElement.variable = variable
    pointerElement.type = type_
    pointer = Pointer(elements=(pointerElement,))

    return pointer


@get_pointer.register(GeppettoModel, str)
def get_pointer_model(model: GeppettoModel, instance_path):
    pointer = Pointer(path=instance_path)

    st = iter(instance_path.split("."))
    lastType = None

    for token in st:
        element = PointerElement()
        v = None
        if lastType is None:
            v = find_instance_variable(get_variable(token), model)
            if v is None:
                #  it's not an instance but it might a library
                library = find_library(model, token)
                if library is not None and st:
                    type_ = next(st)
                    lastType = get_type(model, token + "." + type_)
                    element.type = lastType
                    pointer.elements.add(element)
                    continue
                else:
                    raise GeppettoModelException(token + " is neither an instance variable nor a library id")
        else:
            if isinstance(lastType, CompositeType):
                v = find_variable(get_variable(token), lastType)
            else:
                if isinstance(lastType, ArrayType) and isinstance(lastType.arrayType, CompositeType):
                    v = find_variable(get_variable(token), (lastType).arrayType)
                else:
                    raise GeppettoModelException(
                        lastType.name + " is not of type CompositeType: there can't be nested variables")
        lastType = find_type(get_type_str(token), v)
        element.variable = v
        element.type = lastType
        if isinstance(element.type, ArrayType):
            index = get_index(token)
            if index is not None:
                element.index = get_index(token)
        pointer.elements.add(element)
    return pointer


@multidispatch
def get_type(model, path):
    st = iter(path.split('.'))
    lastType = None
    lastVar = None
    library = None
    for token in st:

        #  token can be a library, a type or a variable
        if lastType is not None:
            if isinstance(lastType, CompositeType):
                lastVar = find_variable(get_variable(token), lastType)
            elif isinstance(lastType, ArrayType) and isinstance(lastType.arrayType, CompositeType):
                lastVar = find_variable(get_variable(token), (lastType).arrayType)
            else:
                raise GeppettoModelException(
                    "{} is not of type CompositeType there can't be nested variables".format(lastType.id))
        elif lastVar is not None:
            lastType = find_type(get_type_str(token), lastVar)
        elif library is not None:
            lastType = find_type(token, library)
        else:
            #  they are all null
            library = find_library(model, token)
            if library is None:
                raise GeppettoModelException("Can't find a type for the path " + path)
    if lastType is not None and lastType.getPath() == path:
        return lastType
    else:
        raise GeppettoModelException("Couldn't find a type for the path " + path)


@get_type.register(str)
def get_type_str(token: str):
    if "(" in token:
        return token[token.find("(") + 1: token.find(")")]
    else:
        return None


@get_type.register(Pointer)
def get_type_pointer(pointer: Pointer):
    return pointer.elements[-1].type


def get_value(model, path, state_variable_type=None):
    st = iter(path.split('.'))
    lastType = None
    lastVar = None
    library = None
    for token in st:
        #  token can be a library, a type or a variable
        if lastType is not None:
            if isinstance(lastType, CompositeType):
                lastVar = find_variable(get_variable(token), lastType)
                lastType = None
            else:
                if isinstance(lastType, ArrayType) and isinstance(lastType.arrayType, CompositeType):
                    lastVar = find_variable(get_variable(token), (lastType).arrayType)
                else:
                    raise GeppettoModelException(
                        "{} is not of type CompositeType there can't be nested variables".format(lastType.id))
        elif lastVar is not None:
            lastType = find_type(get_type_str(token), lastVar)
        elif library is not None:
            lastType = find_type(token, library)
        else:
            #  they are all null
            library = find_library(model, token)
            if library is None:
                raise GeppettoModelException("Can't find a value for the path " + path)
    if lastType is not None and lastType.path == path:
        return lastType.defaultValue
    elif lastVar is not None:
        if state_variable_type is not None:
            try:
                return next(type_value_map.value for type_value_map in lastVar.initialValues if
                            type_value_map.type == state_variable_type)
            except StopIteration as e:
                raise GeppettoModelException("Couldn't find a value for the path " + path) from e
        else:
            return lastVar.initialValues[0].value
    else:
        raise GeppettoModelException("Couldn't find a value for the path " + path)


#
# 	 * @param model
# 	 * @param libraryId
# 	 * @return
# 	 * @throws GeppettoModelException
#
def find_library(model, libraryId):
    return next((library for library in model.libraries if library.id == libraryId), None)


#
# 	 * @param pointer
# 	 * @param pointer2
# 	 * @return true if the two pointers point to the same variables and types
#
@multidispatch
def equals(pointer: Pointer, pointer2: Pointer):
    if not pointer == pointer2:
        if len(pointer.elements) != len(pointer2.elements):
            return False
        for i, pe in enumerate(pointer.elements):
            try:
                pe2 = pointer2.elements[i]
            except:
                return False

            if not pe2 or not equals(pe, pe2):
                return False
    return True


#
# 	 * @param pointer
# 	 * @param pointer2
# 	 * @return
#
@equals.register(PointerElement, PointerElement)
def equals_pointer(pointer: PointerElement, pointer2: PointerElement):
    sameType = pointer.type == pointer2.type or pointer.type == pointer2.type
    sameVar = pointer.variable == pointer2.variable or pointer.variable == pointer2.variable
    sameIndex = pointer.index == pointer2.index
    return sameType and sameVar and sameIndex


#
# 	 * @param pointer
# 	 * @return
#
@multidispatch
def get_variable(pointer: Pointer):
    return pointer.elements[-1].variable


@get_variable.register(str)
def get_variable_str(token: str) -> str:
    if "(" in token:
        return token[0: token.find("(")]
    elif "[" in token:
        return token[0: token.find("[")]
    else:
        return token


def get_geppetto_library(pointer):
    type_ = get_type(pointer)
    while not isinstance(type_.eContainer(), GeppettoLibrary):
        type_ = type_.eContainer().eContainer()
    if (type_.eContainer()).id == "common":
        var = get_variable(pointer)
        type_ = var.eContainer()
        while not isinstance(type_.eContainer(), GeppettoLibrary):
            type_ = type_.eContainer().eContainer()
    return type_.eContainer()


def get_instance_path(variable, type_):
    return variable.id + "(" + type_.id + ")"


@multidispatch
def find_type(type_, variable: Variable) -> Type:
    if type_ is None:
        types = []
        types += variable.anonymousTypes
        types += variable.types
        if len(types) == 1:
            return types[0]
        elif len(types) == 0:
            raise GeppettoModelException("The variable {} has not types".format(variable.id))
        else:
            raise GeppettoModelException(
                "The instance path does not specify a type but more than one type are present for the variable {}".format(
                    variable.id))
    else:
        for t in variable.types:
            if t.id == type_:
                return t
        for t in variable.anonymousTypes:
            if t.id == type_:
                return t
        raise GeppettoModelException("The type {} was not found in the variable {}".format(type_, variable.id))


@find_type.register(str, GeppettoLibrary)
def find_type_library(typeId, library: GeppettoLibrary):
    return next((type_ for type_ in library.types if type_.id == typeId), None)


def find_instance_variable(variablename: str, model: GeppettoModel):
    for v in model.variables:
        if v.id == variablename:
            return v
    return None


def find_variable_from_path(model: GeppettoModel, path: str):
    return get_variable(get_pointer(model, path))


def find_variable(variablename: str, type_: Type) -> Variable:
    for v in type_.variables:
        if v.id == variablename:
            return v
    raise GeppettoModelException("The variable {} was not found in the type {}".format(variablename, type_.id))


def get_index(token):
    if "[" in token:
        return int(token[token.find("[") + 1: token.find("]")])
    else:
        return None


def get_path_without_types(path):
    return path.replaceAll("\\([^)]*\\)", "")



# pylint: disable=redefined-outer-name
import pytest
import responses
from pygeppetto.model import GeppettoLibrary, Variable
from pygeppetto.model.model_access import GeppettoModelAccess
from pygeppetto.model.datasources.datasources import QueryResults, QueryResult
from pygeppetto.model.datasources.datasources import SimpleQuery, CompoundQuery, ProcessQuery
from pygeppetto.model.datasources.datasources import QueryMatchingCriteria, DataSource

from pygeppetto.visitors.data_source_visitors import ExecuteQueryVisitor
from pygeppetto.model.utils.datasource import query_check

from pygeppetto.services.model_interpreter import add_model_interpreter
from pygeppetto.model.types import CompositeVisualType, VisualType
from pygeppetto.services.data_source.neo4j import Neo4jDataSourceService
from pygeppetto.services.data_source_service import QueryProcessor
from .mocks import MockModelInterpreter, neo4j_response as mock_neo4j_response

URL = "http://localhost:7474/db/data/transaction/commit"

class MockQueryProcessor(QueryProcessor):
    def process(self, query: ProcessQuery, data_source: DataSource, variable, results: QueryResults,
                model_access: GeppettoModelAccess) -> QueryResults:
        # lets remove the first result to make sure the processor is applied
        results.results.pop(0)
        return results


def create_result(iden):
    return (iden, f"Result number {iden}")

def create_query_result(iden):
    return QueryResult(values=create_result(iden))

def model_access():
    model_interpreter = MockModelInterpreter()
    model_library = GeppettoLibrary(name='mocklibrary', id='mocklibrary')
    geppetto_model = model_interpreter.create_model(library=model_library)
    add_model_interpreter(model_library.id, model_interpreter)
    return GeppettoModelAccess(geppetto_model)


@pytest.fixture
def visitor():
    return ExecuteQueryVisitor(node_path="visitor", types=(CompositeVisualType(),),
                               geppetto_model_access=model_access(),
                               count_only=False,
                               processing_output_map=None)


def test_visitor_merge_results(visitor):
    header = ["ID", "content"]

    assert visitor.results is None

    initial_results = [create_query_result(i) for i in range(2)]

    query_results = QueryResults(id="initial", header=header, results=initial_results)
    visitor.merge_results(query_results)
    assert len(visitor.results.results) == 2
    assert all([result.values[1] == f"Result number {index}" \
                                for index, result in enumerate(visitor.results.results)])

    new_results = [create_query_result(i+2) for i in range(3)]
    query_results = QueryResults(id="new", header=header, results=new_results)
    visitor.merge_results(query_results)

    assert len(visitor.results.results) == 5
    assert all([result.values[1] == f"Result number {index}" \
                                for index, result in enumerate(visitor.results.results)])

    modified_results = [QueryResult(values=(0, "Result number 0 modified"))]
    query_results = QueryResults(id="modified", header=header, results=modified_results)

    visitor.merge_results(query_results)

    assert len(visitor.results.results) == 5
    assert visitor.results.results[0].values[2] == 'Result number 0 modified'


@responses.activate
def test_simple_query_case(visitor):

    responses.add(responses.POST, URL, json=mock_neo4j_response(), status=200, stream=True)

    v_type = VisualType()
    match_crit = QueryMatchingCriteria(type=(v_type,))
    query = SimpleQuery(query="\"statement\": \"MATCH(n) WHERE id(n)=$ID RETURN id(n) as ID, n;\"",
                        matchingCriteria=(match_crit,))

    DataSource(url=URL, queries=(query,), dataSourceService=Neo4jDataSourceService.__name__)

    visitor.do_switch(query)

    assert len(visitor.results.results) == 2
    assert visitor.results.results[0].values == ['0', '{"title": "The Matrix", "released": 1999}']


@responses.activate
def test_compound_query_case(visitor):
    responses.add(responses.POST, URL, json=mock_neo4j_response(), status=200, stream=True)

    v_type = VisualType()
    match_crit = QueryMatchingCriteria(type=(v_type,))
    query = SimpleQuery(query="\"statement\": \"MATCH(n) RETURN id(n) as ID, n;\"",
                        matchingCriteria=(match_crit,))

    compound_query = CompoundQuery(queryChain=[query, query])

    DataSource(url=URL, queries=(compound_query,),
               dataSourceService=Neo4jDataSourceService.__name__)

    visitor.do_switch(compound_query)

    assert len(visitor.results.results) == 2
    assert visitor.results.results[0].values == ['0', '{"title": "The Matrix", "released": 1999}']


@responses.activate
def test_process_query_case(visitor):
    responses.add(responses.POST, URL, json=mock_neo4j_response(), status=200, stream=True)

    v_type = VisualType()
    match_crit = QueryMatchingCriteria(type=(v_type,))

    query = SimpleQuery(query="\"statement\": \"MATCH(n) RETURN id(n) as ID, n;\"",
                        matchingCriteria=(match_crit,))

    process_query = ProcessQuery(queryProcessorId="MockQueryProcessor", parameters="")

    compound_query = CompoundQuery(queryChain=[query, process_query])

    DataSource(url=URL, queries=(compound_query,), dataSourceService="Neo4jDataSourceService")

    MockQueryProcessor()

    visitor.do_switch(compound_query)

    assert len(visitor.results.results) == 1
    assert visitor.results.results[0].values == ['1', '{"released": 1964, "title": "Keanu Reeves"}']


def test_query_check():
    vtype = VisualType()
    comp_vtype = CompositeVisualType()


    mc1 = QueryMatchingCriteria(type=(vtype,))
    mc2 = QueryMatchingCriteria(type=(comp_vtype,))

    query1 = SimpleQuery(query="dummy query;", matchingCriteria=(mc1,))
    query2 = SimpleQuery(query="dummy query;", matchingCriteria=(mc2,))

    assert not query_check(query=query1, types=(vtype,))
    assert query_check(query=query1, types=(comp_vtype,))
    assert not query_check(query=query2, types=(vtype,))



import os

import pytest

from pyecore.resources import ResourceSet, URI

HERE = os.path.dirname(os.path.abspath(__file__))

@pytest.fixture
def rset():
    return ResourceSet()


def filepath(filename):
    return os.path.join(HERE, 'xmi-data', filename)




def test_read_mediumXMI(rset):
    resource = rset.get_resource(URI(filepath('MediumNet.net.nml.xmi')))
    root = resource.contents[0]
    assert root  # The root exists

@pytest.mark.skip('Too slow')
def test_read_BigXMI(rset):
    resource = rset.get_resource(URI(filepath('BigCA1.net.nml.xmi')))
    root = resource.contents[0]
    assert root  # The root exists

@pytest.mark.skip('Too slow')
def test_read_LargeXMI(rset):
    resource = rset.get_resource(URI(filepath('LargeConns.net.nml.xmi')))
    root = resource.contents[0]
    assert root  # The root exists


def test_readwrite_mediumXMI(tmpdir, rset):
    resource = rset.get_resource(URI(filepath('MediumNet.net.nml.xmi')))
    root = resource.contents[0]
    f = tmpdir.mkdir('pyecore-tmp').join('medium.xmi')
    resource.save(output=URI(str(f)))

@pytest.mark.skip('Too slow')
def test_readwrite_BigXMI(tmpdir, rset):
    resource = rset.get_resource(URI(filepath('BigCA1.net.nml.xmi')))
    root = resource.contents[0]
    f = tmpdir.mkdir('pyecore-tmp').join('big.xmi')
    resource.save(output=URI(str(f)))

@pytest.mark.skip('Too slow')
def test_readwrite_LargeXMI(tmpdir, rset):
    resource = rset.get_resource(URI(filepath('LargeConns.net.nml.xmi')))
    root = resource.contents[0]
    f = tmpdir.mkdir('pyecore-tmp').join('large.xmi')
    resource.save(output=URI(str(f)))


def test_roundtrip_mediumXMI(tmpdir, rset):
    resource = rset.get_resource(URI(filepath('MediumNet.net.nml.xmi')))
    root = resource.contents[0]

    # We change the root name
    root.name = 'mediumTestModel'

    # We serialize the modifications
    f = tmpdir.mkdir('pyecore-tmp').join('medium.xmi')
    resource.save(output=URI(str(f)))

    # We read again the file
    resource = rset.get_resource(URI(str(f)))
    root = resource.contents[0]
    assert root
    assert root.name == 'mediumTestModel'

@pytest.mark.skip('Too slow')
def test_roundtrip_BigXMI(tmpdir, rset):
    resource = rset.get_resource(URI(filepath('BigCA1.net.nml.xmi')))
    root = resource.contents[0]

    # We change the root name
    root.name = 'bigTestModel'

    # We serialize the modifications
    f = tmpdir.mkdir('pyecore-tmp').join('big.xmi')
    resource.save(output=URI(str(f)))

    # We read again the file
    resource = rset.get_resource(URI(str(f)))
    root = resource.contents[0]
    assert root
    assert root.name == 'bigTestModel'

@pytest.mark.skip('Too slow')
def test_roundtrip_LargeXMI(tmpdir, rset):
    resource = rset.get_resource(URI(filepath('LargeConns.net.nml.xmi')))
    root = resource.contents[0]

    # We change the root name
    root.name = 'largeTestModel'

    # We serialize the modifications
    f = tmpdir.mkdir('pyecore-tmp').join('large.xmi')
    resource.save(output=URI(str(f)))

    # We read again the file
    resource = rset.get_resource(URI(str(f)))
    root = resource.contents[0]
    assert root
    assert root.name == 'largeTestModel'



from pyecore.resources import global_registry
from .datasources import getEClassifier, eClassifiers
from .datasources import name, nsURI, nsPrefix, eClass
from .datasources import DataSource, DataSourceLibraryConfiguration, Query, ProcessQuery, SimpleQuery, CompoundQuery, CompoundRefQuery, QueryResults, RunnableQuery, AQueryResult, QueryResult, SerializableQueryResult, QueryMatchingCriteria, BooleanOperator

from ..types import Type
from ..model import StringToStringMap, Tag, GeppettoLibrary

from . import datasources
from .. import model


__all__ = ['DataSource', 'DataSourceLibraryConfiguration', 'Query', 'ProcessQuery', 'SimpleQuery', 'CompoundQuery', 'CompoundRefQuery',
           'QueryResults', 'RunnableQuery', 'AQueryResult', 'QueryResult', 'SerializableQueryResult', 'QueryMatchingCriteria', 'BooleanOperator']

eSubpackages = []
eSuperPackage = model
datasources.eSubpackages = eSubpackages
datasources.eSuperPackage = eSuperPackage


otherClassifiers = [BooleanOperator]

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for classif in eClassifiers.values():
    eClass.eClassifiers.append(classif.eClass)

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)

register_packages = [datasources] + eSubpackages
for pack in register_packages:
    global_registry[pack.nsURI] = pack



from pyecore.ecore import EAttribute, EDouble
from pygeppetto.model import Type, TimeSeries, Pointer, GeppettoModel, Node, Value


# Implement Type.extends_type
def extends_type(self, type):
    return isinstance(self, type.__class__)


Type.extends_type = extends_type

# Fix timeSeries unique attributes check
from pyecore.valuecontainer import EList, EOrderedSet
from pyecore.notification import Notification, Kind


def no_check_eattribute_extend(self, sublist):
    super(EList, self).extend(sublist)
    self.owner.notify(Notification(new=sublist,
                                   feature=self.feature,
                                   kind=Kind.ADD_MANY))
    self.owner._isset.add(self.feature)


TimeSeries.value = value = EAttribute(name='value', eType=EDouble, derived=False, changeable=True, upper=-1,
                                      unique=False)

EList.no_check_eattribute_extend = no_check_eattribute_extend

default_timeseries_init = TimeSeries.__init__


def timeseries_init(self, unit=None, scalingFactor=None, value=None, **kwargs):
    default_timeseries_init(self, unit=unit, scalingFactor=scalingFactor, **kwargs)
    if value:
        self.value.no_check_eattribute_extend(value)


TimeSeries.__init__ = timeseries_init


# Implement method Pointer.get_instance_path
def get_instance_path(self):
    """ generated source for method getInstancePath """
    instance_path = []
    for i, element in enumerate(self.elements):
        instance_path.append(element.variable.id)
        instance_path.append("(" + element.type.id + ")")
        if element.index is not None and element.index > -1:
            instance_path.append("[{0}]".format(element.index))
        if i != len(self.elements) - 1:
            instance_path.append(".")
    return ''.join(instance_path)


Pointer.get_instance_path = Pointer.getInstancePath = get_instance_path


# Implement method Node.getPath
def getPath(self):
    if not (isinstance(self.eContainer(), GeppettoModel)) and isinstance(self.eContainer(), Node):
        container = self.eContainer()
        from pygeppetto.model import Variable  # Must import locally, circular reference otherwise
        if container.eContainer().__class__ == Variable:
            container = container.eContainer()
        return container.getPath() + "." + self.id
    else:
        return self.id


Node.getPath = Node.get_path = getPath



"""Definition of meta model 'model'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *


name = 'model'
nsURI = 'https://raw.githubusercontent.com/openworm/org.geppetto.model/master/src/main/resources/geppettoModel.ecore'
nsPrefix = 'gep'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)
FileFormat = EEnum('FileFormat', literals=['ZIP', 'HDF5'])


class GeppettoModel(EObject, metaclass=MetaEClass):
    """The root of every Geppetto model. This is the configuration of the Geppetto Model, to not be confused with the model Instantiation."""
    id = EAttribute(eType=EString, derived=False, changeable=True)
    name = EAttribute(eType=EString, derived=False, changeable=True)
    variables = EReference(ordered=True, unique=True, containment=True, upper=-1)
    worlds = EReference(ordered=True, unique=True, containment=True, upper=-1)
    libraries = EReference(ordered=True, unique=True, containment=True, upper=-1)
    tags = EReference(ordered=True, unique=True, containment=True, upper=-1)
    dataSources = EReference(ordered=True, unique=True, containment=True, upper=-1)
    queries = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, variables=None, worlds=None, libraries=None, tags=None, id=None, name=None, dataSources=None, queries=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if id is not None:
            self.id = id

        if name is not None:
            self.name = name

        if variables:
            self.variables.extend(variables)

        if worlds:
            self.worlds.extend(worlds)

        if libraries:
            self.libraries.extend(libraries)

        if tags:
            self.tags.extend(tags)

        if dataSources:
            self.dataSources.extend(dataSources)

        if queries:
            self.queries.extend(queries)


class LibraryManager(EObject, metaclass=MetaEClass):

    libraries = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, libraries=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if libraries:
            self.libraries.extend(libraries)


class ExperimentState(EObject, metaclass=MetaEClass):

    experimentId = EAttribute(eType=ELong, derived=False, changeable=True)
    projectId = EAttribute(eType=ELong, derived=False, changeable=True)
    recordedVariables = EReference(ordered=True, unique=True, containment=True, upper=-1)
    setParameters = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, recordedVariables=None, setParameters=None, experimentId=None, projectId=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if experimentId is not None:
            self.experimentId = experimentId

        if projectId is not None:
            self.projectId = projectId

        if recordedVariables:
            self.recordedVariables.extend(recordedVariables)

        if setParameters:
            self.setParameters.extend(setParameters)


class VariableValue(EObject, metaclass=MetaEClass):

    pointer = EReference(ordered=True, unique=True, containment=True)
    value = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, pointer=None, value=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if pointer is not None:
            self.pointer = pointer

        if value is not None:
            self.value = value


class DomainModel(EObject, metaclass=MetaEClass):

    domainModel = EAttribute(eType=EJavaObject, derived=False, changeable=True)
    format = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, domainModel=None, format=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if domainModel is not None:
            self.domainModel = domainModel

        if format is not None:
            self.format = format


class ModelFormat(EObject, metaclass=MetaEClass):

    modelFormat = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, modelFormat=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if modelFormat is not None:
            self.modelFormat = modelFormat


class StringToStringMap(EObject, metaclass=MetaEClass):

    key = EAttribute(eType=EString, derived=False, changeable=True)
    value = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, key=None, value=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if key is not None:
            self.key = key

        if value is not None:
            self.value = value


@abstract
class ISynchable(EObject, metaclass=MetaEClass):

    synched = EAttribute(eType=EBoolean, derived=False, changeable=True)

    def __init__(self, synched=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if synched is not None:
            self.synched = synched


@abstract
class Node(ISynchable):

    id = EAttribute(eType=EString, derived=False, changeable=True)
    name = EAttribute(eType=EString, derived=False, changeable=True)
    tags = EReference(ordered=True, unique=True, containment=False, upper=-1)

    def __init__(self, id=None, name=None, tags=None, **kwargs):

        super().__init__(**kwargs)

        if id is not None:
            self.id = id

        if name is not None:
            self.name = name

        if tags:
            self.tags.extend(tags)

    def getPath(self):

        raise NotImplementedError('operation getPath(...) not yet implemented')


class Tag(ISynchable):

    name = EAttribute(eType=EString, derived=False, changeable=True)
    tags = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, tags=None, name=None, **kwargs):

        super().__init__(**kwargs)

        if name is not None:
            self.name = name

        if tags:
            self.tags.extend(tags)


class ExternalDomainModel(DomainModel):

    fileFormat = EAttribute(eType=FileFormat, derived=False,
                            changeable=True, default_value=FileFormat.ZIP)

    def __init__(self, fileFormat=None, **kwargs):

        super().__init__(**kwargs)

        if fileFormat is not None:
            self.fileFormat = fileFormat


class GeppettoLibrary(Node):

    types = EReference(ordered=True, unique=True, containment=True, upper=-1)
    sharedTypes = EReference(ordered=True, unique=True, containment=False, upper=-1)

    def __init__(self, types=None, sharedTypes=None, **kwargs):

        super().__init__(**kwargs)

        if types:
            self.types.extend(types)

        if sharedTypes:
            self.sharedTypes.extend(sharedTypes)

    def getTypeById(self):

        raise NotImplementedError('operation getTypeById(...) not yet implemented')


class World(Node):

    variables = EReference(ordered=True, unique=True, containment=True, upper=-1)
    instances = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, variables=None, instances=None, **kwargs):

        super().__init__(**kwargs)

        if variables:
            self.variables.extend(variables)

        if instances:
            self.instances.extend(instances)



import json

from pygeppetto.utils import Singleton

try:
    from types import SimpleNamespace as Namespace
except ImportError:
    # Python 2.x fallback
    from argparse import Namespace


class GeppettoDataManager(object, metaclass=Singleton):
    """ generated source for interface IGeppettoDataManager """

    def __init__(self):
        self.projects = {}

    def getName(self):
        raise NotImplemented('Operation not defined yet in default data manager')

    def isDefault(self):
        return True

    def getUserByLogin(self, login):
        raise NotImplemented('Operation not defined yet in default data manager')

    def getUserGroupById(self, id):
        raise NotImplemented('Operation not defined yet in default data manager')

    def getGeppettoProjectById(self, project_id):
        return self.projects.get(project_id, None)

    def getAllUsers(self):
        raise NotImplemented('Operation not defined yet in default data manager')

    def getAllGeppettoProjects(self):
        raise NotImplemented('Operation not defined yet in default data manager')

    def getGeppettoProjectsForUser(self, login):
        raise NotImplemented('Operation not defined yet in default data manager')

    def getProjectFromJson(self, json_str):
        project = json.loads(json_str, object_hook=lambda d: Namespace(**d))
        project.volatile = True
        project.id = hash(json_str)
        self.projects[project.id] = project
        return project

    def getExperimentsForProject(self, projectId):
        raise NotImplemented('Operation not defined yet in default data manager')

    def newSimulationResult(self, parameterPath, results, format):
        raise NotImplemented('Operation not defined yet in default data manager')

    def addWatchedVariable(self, found, instancePath):
        raise NotImplemented('Operation not defined yet in default data manager')

    def newPersistedData(self, url, type_):
        raise NotImplemented('Operation not defined yet in default data manager')

    def newParameter(self, parameterPath, value):
        raise NotImplemented('Operation not defined yet in default data manager')

    def newExperiment(self, name, description, project):
        raise NotImplemented('Operation not defined yet in default data manager')

    def newView(self, view, project):
        raise NotImplemented('Operation not defined yet in default data manager')

    def newView_0(self, view, experiment):
        raise NotImplemented('Operation not defined yet in default data manager')

    def newUser(self, name, password, persistent, group):
        raise NotImplemented('Operation not defined yet in default data manager')

    def newUserGroup(self, name, privileges, spaceAllowance, timeAllowance):
        raise NotImplemented('Operation not defined yet in default data manager')

    def updateUser(self, user, password):
        raise NotImplemented('Operation not defined yet in default data manager')

    def newAspectConfiguration(self, experiment, instancePath, simulatorConfiguration):
        raise NotImplemented('Operation not defined yet in default data manager')

    def newSimulatorConfiguration(self, simulator, conversionService, timestep, length, parameters):
        raise NotImplemented('Operation not defined yet in default data manager')

    def addGeppettoProject(self, project, user):
        raise NotImplemented('Operation not defined yet in default data manager')

    def makeGeppettoProjectPublic(self, projectId, isPublic):
        raise NotImplemented('Operation not defined yet in default data manager')

    def deleteGeppettoProject(self, id, user):
        raise NotImplemented('Operation not defined yet in default data manager')

    def deleteExperiment(self, experiment):
        raise NotImplemented('Operation not defined yet in default data manager')

    def clearWatchedVariables(self, aspectConfig):
        raise NotImplemented('Operation not defined yet in default data manager')

    def saveEntity(self, entity):
        raise NotImplemented('Operation not defined yet in default data manager')

    def saveEntity_0(self, entity):
        raise NotImplemented('Operation not defined yet in default data manager')

    def saveEntity_1(self, entity):
        raise NotImplemented('Operation not defined yet in default data manager')

    def cloneExperiment(self, name, description, project, originalExperiment):
        raise NotImplemented('Operation not defined yet in default data manager')

    def get_project_from_url(self, urlString):
        with open(urlString) as json_file:
            return self.getProjectFromJson(json_file.read())


class DataManagerHelper:
    __data_manager = GeppettoDataManager()

    def __init__(self):
        raise NotImplementedError("DataManagerHelper cannot be instantiated. Call class methods directly instead")

    @classmethod
    def getDataManager(cls) -> GeppettoDataManager:
        return cls.__data_manager

    @classmethod
    def setDataManager(cls, data_manager: GeppettoDataManager):
        cls.__data_manager = data_manager


set_data_manager = DataManagerHelper.setDataManager
get_data_manager = DataManagerHelper.getDataManager



import sys

InboundMessages = sys.modules[__name__]

GEPPETTO_VERSION = "geppetto_version"

USER_PRIVILEGES = "user_privileges"

NOTIFY_USER = "notify_user"

GET_SCRIPT = "get_script"

GET_DATA_SOURCE_RESULTS = "get_data_source_results"

LOAD_PROJECT_FROM_URL = "load_project_from_url"

LOAD_PROJECT_FROM_ID = "load_project_from_id"

LOAD_PROJECT_FROM_CONTENT = "load_project_from_content"

SAVE_PROJECT_PROPERTIES = "save_project_properties"

PERSIST_PROJECT = "persist_project"

MAKE_PROJECT_PUBLIC = "make_project_public"

DOWNLOAD_PROJECT = "download_project"

NEW_EXPERIMENT = "new_experiment"

NEW_EXPERIMENT_BATCH = "new_experiment_batch"

CLONE_EXPERIMENT = "clone_experiment"

LOAD_EXPERIMENT = "load_experiment"

SAVE_EXPERIMENT_PROPERTIES = "save_experiment_properties"

DELETE_EXPERIMENT = "delete_experiment"

GET_EXPERIMENT_STATE = "get_experiment_state"

EXPERIMENT_STATUS = "experiment_status"

RUN_EXPERIMENT = "run_experiment"

SET_WATCHED_VARIABLES = "set_watched_variables"

GET_WATCH = "get_watch"

CLEAR_WATCHED_VARIABLES = "clear_watch"

SET_PARAMETERS = "set_parameters"

SET_EXPERIMENT_VIEW = "set_experiment_view"

LINK_DROPBOX = "link_dropbox"

UNLINK_DROPBOX = "unlink_dropbox"

GET_DROPBOX_TOKEN = "get_dropbox_token"

UPLOAD_MODEL = "upload_model"

UPLOAD_RESULTS = "upload_results"

GET_SUPPORTED_OUTPUTS = "get_supported_outputs"

DOWNLOAD_MODEL = "download_model"

DOWNLOAD_RESULTS = "download_results"

FETCH_VARIABLE = "fetch_variable"

RESOLVE_IMPORT_TYPE = "resolve_import_type"

RESOLVE_IMPORT_VALUE = "resolve_import_value"

RUN_QUERY = "run_query"

RUN_QUERY_COUNT = "run_query_count"

FETCH = 'fetch'



import json
import logging

from pygeppetto.model import QueryResults
from pygeppetto.model.datasources import QueryResult
from pygeppetto.services.data_source_service import DataSourceService

NEO4J_ERROR_PARSING_RESPONSE = "Error parsing Neo4j response"
NEO4J_ERROR_MISSING_COLUMN_KEY_IN_RESPONSE = "Missing `columns` key in Neo4j response."
NEO4J_ERROR_MISSING_ROW_KEY_IN_RESPONSE = "Missing `row` or `meta` key in Neo4j response."
NEO4J_ERROR_MISSING_ERROR_KEY_IN_RESPONSE = 'Neo4j: missing "error" key in Neo4j response.'

NEO4J_ROW_KEY = "row"
NEO4J_META_KEY = "meta"
NEO4J_DATA_KEY = "data"
NEO4J_ERRORS_KEY = "errors"
NEO4J_RESULTS_KEY = "results"
NEO4J_COLUMNS_KEY = "columns"

TEMPLATE = '''{
  "statements":[
    {
      $QUERY,
      "resultDataContents":[
        "row"
      ],
      "includeStats":false
    }
  ]
}'''


class Neo4jDataSourceService(DataSourceService):

    def get_connection_type(self):
        return "POST"

    def get_template(self):
        return TEMPLATE

    def process_response(self, response):
        query_results = QueryResults(header=[], results=[])

        for chunk in response:
            try:
                chunk = json.loads(chunk)
            except:
                raise Exception(NEO4J_ERROR_PARSING_RESPONSE)

            if NEO4J_ERRORS_KEY in chunk and not chunk[NEO4J_ERRORS_KEY]:
                for container in chunk[NEO4J_RESULTS_KEY]:
                    if NEO4J_COLUMNS_KEY in container:
                        query_results.header.update(container[NEO4J_COLUMNS_KEY])
                    else:
                        logging.error(NEO4J_ERROR_MISSING_COLUMN_KEY_IN_RESPONSE)
                        break

                    for data in container[NEO4J_DATA_KEY]:
                        if NEO4J_ROW_KEY in data and NEO4J_META_KEY in data:
                            # can't store r as dict because it is unhashable
                            row = [json.dumps(r) for r in data[NEO4J_ROW_KEY]]
                            query_results.results.append(QueryResult(values=row))
                        else:
                            logging.error(NEO4J_ERROR_MISSING_ROW_KEY_IN_RESPONSE)
                            break
            else:
                if not NEO4J_ERRORS_KEY in chunk:
                    logging.error(NEO4J_ERROR_MISSING_ERROR_KEY_IN_RESPONSE)
                else:
                    for error in chunk[NEO4J_ERRORS_KEY]:
                        logging.error('Neo4j: error_code %s, error_message: %s.',
                                      error["code"], error["message"])

        return query_results



from pyecore.resources import global_registry
from .values import getEClassifier, eClassifiers
from .values import name, nsURI, nsPrefix, eClass
from .values import Value, Composite, StringToValueMap, Quantity, PhysicalQuantity, Unit, TimeSeries, MDTimeSeries, MetadataValue, Text, URL, HTML, Pointer, PointerElement, Point, Dynamics, FunctionPlot, Function, Argument, Expression, VisualValue, Collada, OBJ, Sphere, Cylinder, Particles, SkeletonAnimation, SkeletonTransformation, VisualGroupElement, VisualGroup, Connection, Connectivity, ArrayElement, ArrayValue, Image, ImageFormat, ImportValue, Metadata, JSON, GenericArray, StringArray, IntArray, DoubleArray, AArrayValue

from ..types import Type
from ..variables import Variable
from ..model import Tag

from . import values
from .. import model


__all__ = ['Value', 'Composite', 'StringToValueMap', 'Quantity', 'PhysicalQuantity', 'Unit', 'TimeSeries', 'MDTimeSeries', 'MetadataValue', 'Text', 'URL', 'HTML', 'Pointer', 'PointerElement', 'Point', 'Dynamics', 'FunctionPlot', 'Function', 'Argument', 'Expression', 'VisualValue', 'Collada', 'OBJ', 'Sphere',
           'Cylinder', 'Particles', 'SkeletonAnimation', 'SkeletonTransformation', 'VisualGroupElement', 'VisualGroup', 'Connection', 'Connectivity', 'ArrayElement', 'ArrayValue', 'Image', 'ImageFormat', 'ImportValue', 'Metadata', 'JSON', 'GenericArray', 'StringArray', 'IntArray', 'DoubleArray', 'AArrayValue']

eSubpackages = []
eSuperPackage = model
values.eSubpackages = eSubpackages
values.eSuperPackage = eSuperPackage


otherClassifiers = [Connectivity, ImageFormat]

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for classif in eClassifiers.values():
    eClass.eClassifiers.append(classif.eClass)

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)

register_packages = [values] + eSubpackages
for pack in register_packages:
    global_registry[pack.nsURI] = pack



import airspeed

def process_template(template, **properties):
    target = template
    tpl = airspeed.Template(target)
    result = tpl.merge(properties)
    if '$' in target and result != template:
        return process_template(result, **properties)
    return result



import abc
import functools
from enum import Enum, unique

from pygeppetto.constants import UserPrivileges, ExperimentStatus
from pygeppetto.managers.runtime_project import RuntimeProject

from pygeppetto.utils import Singleton

from .experiment_run_manager import ExperimentRunManager

# Creates a Python 2 and 3 compatible base class
ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})

from pygeppetto.model.exceptions import GeppettoAccessException, GeppettoExecutionException, ModelInterpreterException
import logging


@unique
class Scope(Enum):
    RUN = 0
    CONNECTION = 1




def ensure(rights=None, not_scope=None, message="perform the required action"):
    def inner_user_needs_rights(func):
        @functools.wraps(func)
        def manager_function(self, *args, **kwargs):
            has_wrong_scope = not_scope is None or not_scope is not self.scope
            for right in rights:
                if has_wrong_scope and right not in self.user.group.privileges:
                    raise GeppettoAccessException("Insufficient access rights to "
                                                  + message)
            return func(self, *args, **kwargs)

        return manager_function

    return inner_user_needs_rights


class GeppettoManager(object, metaclass=Singleton):
    def __init__(self, manager=None):
        self.opened_projects = {}
        if manager:
            self.opened_projects.update(manager.opened_projects)
        self.scope = Scope.CONNECTION
        self._user = None

    def is_project_open(self, project):
        return project in self.opened_projects

    def is_user_project(self, project):
        project_id = project
        if hasattr(project, 'id'):
            project_id = project.id
        if self.user:
            return any(p.id == project_id for p in self.user.projects)
        return False

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, value):
        if self._user is not None:
            message = """A GeppettoManager is being reused, an user was already set and the user attribute is beeing set again.
Current user: {}, attempted new user: {}""".format(self._user.name, value.name)
            raise GeppettoExecutionException(message)
        self._user = value

    def load_project(self, project):
        if self.scope is not Scope.RUN:
            if self.user and UserPrivileges.READ_PROJECT not in self.user.group.privileges:
                raise GeppettoAccessException("Insufficient access rights to"
                                              "load project")
            is_user_project = (project.volatile or project.is_public
                               or self.is_user_project(project.id))
            if not is_user_project:
                raise GeppettoAccessException('Project not found for the '
                                              'current user')

        if self.is_project_open(project):
            logging.warning('Project was already opened: {}'.format(project))
            del self.opened_projects[project]
            # return self.get_runtime_project(project)
            # raise GeppettoExecutionException('Cannot load two instances of the same project')
        runtime = RuntimeProject(project, self)
        self.opened_projects[project] = runtime
        return runtime

    def close_project(self, project):
        if not self.is_project_open(project):
            raise GeppettoExecutionException('A project without a runtime '
                                             'project cannot be closed')
        # try:
        #     PathConfig.delete_project_tmp_folder(self.scope, project.id)
        # except Exception:
        #     raise GeppettoExecutionException()
        self.opened_projects[project].release()
        del self.opened_projects[project]

    def get_runtime_project(self, project) -> RuntimeProject:
        if not self.is_project_open(project):
            try:
                return self.load_project(project)
            except Exception as e:
                raise GeppettoExecutionException()
        return self.opened_projects[project]

    @ensure(not_scope=Scope.RUN, rights=[UserPrivileges.READ_PROJECT],
            message='load experiment')
    def load_experiment(self, experiment):
        project = experiment.parent_project
        if not self.is_project_open(project) or self.opened_projects[project] is None:
            raise GeppettoExecutionException('Cannot load an experiment for '
                                             'a project that was not loaded')

        runtime_project = self.get_runtime_project(project)
        try:
            runtime_project.open_experiment(experiment)
        except Exception as e:
            raise GeppettoExecutionException(e)

        return runtime_project[experiment].state

    @ensure(not_scope=Scope.RUN, rights=[UserPrivileges.RUN_EXPERIMENT],
            message='run experiment')
    def run_experiment(self, experiment):
        if experiment.status is ExperimentStatus.DESIGN:
            ExperimentRunManager.queueExperiment(self.user, experiment)
        else:
            raise GeppettoExecutionException('Cannot run an experiment whose '
                                             'status is not design')

    @ensure(rights=[UserPrivileges.READ_PROJECT], message='play experiment')
    def get_experiment_state(self, experiment, variables):
        if experiment.status is ExperimentStatus.COMPLETED:
            project = experiment.parent_project
            url_base = project.url_base
            return self.get_runtime_project(project)[experiment].get_experiment_state(variables, url_base)
        else:
            raise GeppettoExecutionException('Cannot play an experiment whose status is not completed')

    # Not yet tested
    @ensure(rights=[UserPrivileges.WRITE_PROJECT], message='create new experiment')
    def new_experiment(self, project):
        experiment = None  # must implement DataManagerHelper
        try:
            self.get_runtime_project(project).populate_new_experiment(experiment)
        except Exception as e:
            raise GeppettoExecutionException(e)
        return experiment

    # @ensure(rights=[UserPrivileges.WRITE_PROJECT], message='import value')
    def resolve_import_value(self, path, geppetto_project, experiment=None):
        return self.get_runtime_project(geppetto_project).resolve_import_value(path)

    # @ensure(rights=[UserPrivileges.WRITE_PROJECT], message='import type')
    def resolve_import_type(self, type_paths, geppetto_project):
        return self.get_runtime_project(geppetto_project).resolve_import_type(type_paths)

    def run_query(self, runnable_queries, geppetto_project):
        return self.get_runtime_project(geppetto_project).run_query(runnable_queries)

    def fetch_variable(self, data_source_id, variable_ids, geppetto_project):
        return self.get_runtime_project(geppetto_project).fetch_variable(data_source_id, variable_ids)

    def fetch(self, data_source_id, variable_ids, instance_ids, geppetto_project):
        return self.get_runtime_project(geppetto_project).fetch(data_source_id, variable_ids, instance_ids)



import pkgutil
import importlib
import requests


from pyecore.ecore import EObject, EClass, EReference


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def clone(eobject, skip_containment=False, skip_refs=False):
    if not isinstance(eobject, EObject) and not isinstance(eobject, type):
        return eobject
    if isinstance(eobject, type):
        eobject = eobject.eClass

    def create_instance(original):
        if isinstance(original, type):
            original = original.eClass
            cloned_object = original.eClass(original.name)
        elif original.eClass is EClass.eClass:
            cloned_object = original.eClass(original.name)
        else:
            cloned_object = original.eClass()
        return cloned_object

    def first_setup(original, cloned):
        references = []
        for feature in original._isset:
            if isinstance(feature, EReference):
                if skip_refs:
                    continue
                opposite = feature.eOpposite
                if opposite and (opposite.containment or opposite in references):
                    continue
                if skip_containment and feature.containment:
                    continue
                references.append(feature)
                continue
            if feature.many:
                cloned.eGet(feature).extend(original.eGet(feature))
            else:
                cloned.eSet(feature, original.eGet(feature))
        return references

    created_elements = {}
    clone_refs = {}
    all_objects = [eobject]
    if not skip_containment:
        all_objects.extend(eobject.eAllContents())

    for e_obj in all_objects:
        cloned = create_instance(e_obj)
        created_elements[e_obj] = cloned
        clone_refs[e_obj] = (cloned, first_setup(e_obj, cloned))

    if skip_refs:
        return created_elements.get(eobject)

    for original, (cloned, refs) in clone_refs.items():
        for feature in refs:
            if feature.many:
                eobjs = [created_elements.get(x, x) for x in original.eGet(feature)]
                cloned.eGet(feature).extend(eobjs)
            else:
                obj = original.eGet(feature)
                cloned.eSet(feature, created_elements.get(obj, obj))
    return created_elements.get(eobject)


def import_submodules(package, recursive=True):
    """ Import all submodules of a module, recursively, including subpackages

    :param package: package (name or actual module)
    :type package: str | module
    :rtype: dict[str, types.ModuleType]
    """
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = {}
    for _, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + '.' + name
        results[full_name] = importlib.import_module(full_name)
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results


def stream_requests(url, params, method="GET", chunk_size=8192):
    processed_output = []
    kwargs = {"url": url, "stream": True}

    if method == "POST":
        method = requests.post
        kwargs["json"] = params

    elif method == "GET":
        method = requests.get
        kwargs["params"] = params

    with method(**kwargs) as result:
        for chunk in result.iter_lines(chunk_size=chunk_size, delimiter=b'\n'):
            if chunk:
                processed_output.append(chunk.decode("utf-8"))

    return processed_output



import pytest
from pygeppetto.model.utils import template

def test_template_processor():
    query = "MATCH(n) WHERE id(n)=$ID RETURN id(n) as ID, n;"

    output = template.process_template(template='{"statement":"$QUERY"}',
                                       ID="visitor", QUERY=query)

    assert output == "{\"statement\":\"MATCH(n) WHERE id(n)=visitor RETURN id(n) as ID, n;\"}"




try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


def getSVG(fig):
    if isinstance(fig, list):
        svgs = []
        for f in fig:
            svgs.append(getSingleSVG(f))
        return svgs
        # Could be written:
        # return [getSingleSVG(f) for f in fig]
    return getSingleSVG(fig)


def getSingleSVG(fig):
    imgdata = StringIO()
    fig.savefig(imgdata, format='svg', transparent=True)
    imgdata.seek(0)  # rewind the data
    svg_dta = imgdata.getvalue()  # this is svg data
    return svg_dta




"""Definition of meta model 'variables'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *
from ..model import Node, ISynchable


name = 'variables'
nsURI = 'https://raw.githubusercontent.com/openworm/org.geppetto.model/master/src/main/resources/geppettoModel.ecore#//variables'
nsPrefix = 'gep'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)


class TypeToValueMap(EObject, metaclass=MetaEClass):

    key = EReference(ordered=True, unique=True, containment=False)
    value = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, key=None, value=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if key is not None:
            self.key = key

        if value is not None:
            self.value = value


class Variable(Node):

    static = EAttribute(eType=EBoolean, derived=False, changeable=True)
    anonymousTypes = EReference(ordered=True, unique=True, containment=True, upper=-1)
    types = EReference(ordered=True, unique=True, containment=False, upper=-1)
    initialValues = EReference(ordered=True, unique=True, containment=True, upper=-1)
    position = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, anonymousTypes=None, types=None, initialValues=None, static=None, position=None, **kwargs):

        super().__init__(**kwargs)

        if static is not None:
            self.static = static

        if anonymousTypes:
            self.anonymousTypes.extend(anonymousTypes)

        if types:
            self.types.extend(types)

        if initialValues:
            self.initialValues.extend(initialValues)

        if position is not None:
            self.position = position



"""Definition of meta model 'types'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *
from ..model import ISynchable, Node


name = 'types'
nsURI = 'https://raw.githubusercontent.com/openworm/org.geppetto.model/master/src/main/resources/geppettoModel.ecore#//types'
nsPrefix = 'gep'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)


@abstract
class Type(Node):

    abstract = EAttribute(eType=EBoolean, derived=False, changeable=True)
    superType = EReference(ordered=True, unique=True, containment=False, upper=-1)
    visualType = EReference(ordered=True, unique=True, containment=False)
    referencedVariables = EReference(ordered=True, unique=True, containment=False, upper=-1)
    domainModel = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, superType=None, abstract=None, visualType=None, referencedVariables=None, domainModel=None, **kwargs):

        super().__init__(**kwargs)

        if abstract is not None:
            self.abstract = abstract

        if superType:
            self.superType.extend(superType)

        if visualType is not None:
            self.visualType = visualType

        if referencedVariables:
            self.referencedVariables.extend(referencedVariables)

        if domainModel is not None:
            self.domainModel = domainModel

    def getDefaultValue(self):

        raise NotImplementedError('operation getDefaultValue(...) not yet implemented')

    def extendsType(self, type=None):

        raise NotImplementedError('operation extendsType(...) not yet implemented')


class VisualType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class ImportType(Type):

    url = EAttribute(eType=EString, derived=False, changeable=True)
    referenceURL = EAttribute(eType=EString, derived=False, changeable=True)
    modelInterpreterId = EAttribute(eType=EString, derived=False, changeable=True)
    autoresolve = EAttribute(eType=EBoolean, derived=False, changeable=True, default_value=True)

    def __init__(self, url=None, referenceURL=None, modelInterpreterId=None, autoresolve=None, **kwargs):

        super().__init__(**kwargs)

        if url is not None:
            self.url = url

        if referenceURL is not None:
            self.referenceURL = referenceURL

        if modelInterpreterId is not None:
            self.modelInterpreterId = modelInterpreterId

        if autoresolve is not None:
            self.autoresolve = autoresolve


class CompositeType(Type):

    variables = EReference(ordered=True, unique=True, containment=True, upper=-1)
    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, variables=None, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if variables:
            self.variables.extend(variables)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class PointerType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class QuantityType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class ParameterType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class StateVariableType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class DynamicsType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class ArgumentType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class ExpressionType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class HTMLType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class JSONType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class TextType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class URLType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class PointType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class ArrayType(Type):

    size = EAttribute(eType=EInt, derived=False, changeable=True)
    arrayType = EReference(ordered=True, unique=True, containment=False)
    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, size=None, arrayType=None, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if size is not None:
            self.size = size

        if arrayType is not None:
            self.arrayType = arrayType

        if defaultValue is not None:
            self.defaultValue = defaultValue


class ConnectionType(Type):

    variables = EReference(ordered=True, unique=True, containment=True, upper=-1)
    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, variables=None, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if variables:
            self.variables.extend(variables)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class SimpleType(Type):

    def __init__(self, **kwargs):

        super().__init__(**kwargs)


class ImageType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class SimpleArrayType(Type):

    defaultValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, defaultValue=None, **kwargs):

        super().__init__(**kwargs)

        if defaultValue is not None:
            self.defaultValue = defaultValue


class MetadataType(Type):

    def __init__(self, **kwargs):

        super().__init__(**kwargs)


class CompositeVisualType(VisualType):

    variables = EReference(ordered=True, unique=True, containment=True, upper=-1)
    visualGroups = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, variables=None, visualGroups=None, **kwargs):

        super().__init__(**kwargs)

        if variables:
            self.variables.extend(variables)

        if visualGroups:
            self.visualGroups.extend(visualGroups)



import json
import os

from pyecore.resources import ResourceSet, URI
from pygeppetto.model.model_factory import GeppettoModelFactory, SharedLibraryManager
from pygeppetto.model.model_serializer import GeppettoModelSerializer as GeppettoSerializer


def test_serializer():
    #  Initialize the factory and the resource set
    res_set = ResourceSet()
    #  How to read
    resource = res_set.get_resource(URI(os.path.join(os.path.dirname(__file__), "xmi-data/GeppettoModelTest.xmi")))

    geppetto_model = resource.contents[0]

    with open(os.path.join(os.path.dirname(__file__), "json-data/test.json")) as testfile:
        json_resource = testfile.read()
    expected_json = json.loads(json_resource)

    # FIXME Empty strings attr are converted to None and the serializer removes the attr (( from: pyecore -> xmi.py -> _init_modelroot))
    geppetto_model.id = ''
    geppetto_model.name = ''
    geppetto_model.variables[0].id = ''
    geppetto_model.libraries[0].id = ''
    geppetto_model.libraries[0].types[0].id = ''

    assert not geppetto_model.variables[0].synched

    # First serialization: the object is serialized without any sync activated.
    produced_json = temporary_fix(json.loads(GeppettoSerializer.serialize(geppetto_model)))
    assert produced_json == expected_json

    assert not geppetto_model.variables[0].synched

    stringify_sync_model = '{"eClass":"GeppettoModel","id":"","name":"","variables":[{"synched":true}],"libraries":[{"synched":true}]}'
    expected_json_synched = json.loads(stringify_sync_model)

    # Second serialization: we activate the sync but we expect the same result.
    produced_json = temporary_fix(json.loads(GeppettoSerializer.serialize(geppetto_model, True)))
    assert produced_json == expected_json
    assert geppetto_model.variables[0].synched
    # Third serialization: we activate the sync and now we find everything on sync.
    produced_json = json.loads(GeppettoSerializer.serialize(geppetto_model, True))
    assert produced_json == expected_json_synched

    # Fourth serialization: the object is synched, but we want the full serialization anyway
    produced_json = temporary_fix(json.loads(GeppettoSerializer.serialize(geppetto_model)))
    assert geppetto_model.variables[0].synched
    assert produced_json == expected_json


def test_references():
    res_set = ResourceSet()
    resource = res_set.get_resource(URI(os.path.join(os.path.dirname(__file__), "xmi-data/GeppettoModelTest.xmi")))
    model1 = resource.contents[0]
    common_library_1 = SharedLibraryManager.get_shared_common_library()
    model1.libraries.append(common_library_1)
    res_set = ResourceSet()
    resource = res_set.get_resource(URI(os.path.join(os.path.dirname(__file__), "xmi-data/GeppettoModelTest.xmi")))
    model2 = resource.contents[0]
    common_library_2 = SharedLibraryManager.get_shared_common_library()
    model2.libraries.append(common_library_2)

    factory = GeppettoModelFactory(common_library_1)

    variable = factory.create_state_variable('0')
    model1.variables.append(variable)
    model1.libraries.append(factory.geppetto_common_library)

    factory2 = GeppettoModelFactory(common_library_2)
    variable = factory2.create_state_variable('0')
    model2.variables.append(variable)

    serialized1 = json.loads(GeppettoSerializer.serialize(model1, False))
    serialized2 = json.loads(GeppettoSerializer.serialize(model2, False))
    assert serialized1 == serialized2

    assert json.loads(GeppettoSerializer.serialize(model1, True)) == json.loads(
        GeppettoSerializer.serialize(model2, True))
    assert json.loads(GeppettoSerializer.serialize(model1, True)) == json.loads(
        GeppettoSerializer.serialize(model2, True))
    assert json.loads(GeppettoSerializer.serialize(model1, False)) == json.loads(
        GeppettoSerializer.serialize(model2, False))


# FIXME Serialization is slightly different from expected
def temporary_fix(produced_json):
    del produced_json['variables'][0]['types'][0]['eClass']
    del produced_json['libraries'][0]['types'][0]['referencedVariables']
    produced_json['variables'][0]['static'] = False
    produced_json['libraries'][0]['types'][0]['abstract'] = False
    produced_json['libraries'][0]['types'][0]['autoresolve'] = True
    if 'synched' in produced_json['libraries'][0].keys():
        del produced_json['libraries'][0]['synched']
        del produced_json['libraries'][0]['types'][0]['synched']
        del produced_json['variables'][0]['synched']
    return produced_json




try:
    from BytesIO import BytesIO
except ImportError:
    from io import BytesIO
from pyecore.resources import URI


class BytesURI(URI):

    def __init__(self, uri, text=None):
        super(BytesURI, self).__init__(uri)
        if text is not None:
            self.__stream = BytesIO(text)

    def getvalue(self):
        return self.__stream.getvalue()

    def create_instream(self):
        return self.__stream

    def create_outstream(self):
        self.__stream = BytesIO()
        return self.__stream



import json

import pytest
from pyecore.ecore import EString
from pyecore.resources import ResourceSet, URI
from pygeppetto.model import TypeToValueMap, Variable
from pygeppetto.model.types import *
from pygeppetto.model.values import *
from pygeppetto.model.model_factory import GeppettoModelFactory


@pytest.fixture()
def model_factory():
    return GeppettoModelFactory()


def test_array_nested_type(model_factory):
    level1_type = CompositeType(id='level1type')
    level1_type = level1_type

    level2_type = CompositeType(id='level2type')
    level2_type = level2_type
    level2_type.variables.append(model_factory.create_text_variable(id='l3', text='default text'))

    vl2 = Variable(id='l2', name='l2', types=(level2_type,))
    level1_type.variables.append(vl2)

    vl1value = Composite()

    vl2value = Composite()
    vl2value.value.append(StringToValueMap('l3', Text('my_text')))

    vl1value.value.append(StringToValueMap('l2', vl2value))

    array_value = ArrayValue()
    array_value.elements.append(ArrayElement(index=0, initialValue=vl1value))

    assert array_value.elements[0].initialValue is not None

    referencing_type = ArrayType(defaultValue=ArrayValue(), arrayType=level1_type)
    v = Variable(id='list', types=(referencing_type,))
    v.initialValues.append(TypeToValueMap(referencing_type, array_value))


def test_create_cylinder(model_factory):
    var = model_factory.create_cylinder('c1')

    assert type(var.types[0]) == VisualType
    assert var.id == 'c1'
    c = var.initialValues[0].value

    assert c.bottomRadius == 1.0
    assert c.topRadius == 1.0

    var = model_factory.create_cylinder('c23', 2.0, 3.0)
    c = var.initialValues[0].value
    assert var.id == 'c23'
    assert c.bottomRadius == 2.0
    assert c.topRadius == 3.0


def test_create_sphere(model_factory):
    var = model_factory.create_sphere('s1')

    assert type(var.types[0]) == VisualType
    assert var.id == 's1'
    c = var.initialValues[0].value

    assert c.radius == 1.0
    assert type(c.position) == Point

    var = model_factory.create_sphere('s23', 2.0, Point(1.0, 1.0, 1.0))
    c = var.initialValues[0].value
    assert var.id == 's23'
    assert c.radius == 2.0


def test_create_state_variable(model_factory):
    var = model_factory.create_state_variable(id='aname', initialValue=Quantity(value=10.0))
    assert var.id == 'aname'
    assert var.initialValues[0].value.value == 10
    assert type(var.types[0]) == StateVariableType


def test_create_time_series_variable(model_factory):
    var = model_factory.create_time_series_variable('aname', values=[1, 2.0], unit='mV')

    assert var.id == 'aname'
    assert type(var.types[0]) == StateVariableType

    ts = var.initialValues[0].value
    assert ts.value[0] == 1
    assert type(ts) == TimeSeries


def test_create_json(model_factory):
    var = model_factory.create_JSON_variable('aname', {'a': 1, 'b': 2})

    assert var.id == 'aname'
    assert type(var.types[0]) == JSONType

    v = json.loads(var.initialValues[0].value.json)
    assert v['a'] == 1


def test_create_parameter_variable(model_factory):
    var = model_factory.create_parameter_variable('aname')

    assert var.id == 'aname'
    assert type(var.types[0]) == ParameterType


def test_create_dynamics_variable(model_factory):
    var = model_factory.create_dynamics_variable('aname')

    assert var.id == 'aname'
    assert type(var.types[0]) == DynamicsType


def test_create_html_variable(model_factory):
    var = model_factory.create_html_variable('aname', '<p>Text</p>')

    assert var.id == 'aname'
    assert type(var.types[0]) == HTMLType

    value = var.initialValues[0].value
    assert type(value) == HTML
    assert value.html == '<p>Text</p>'


def test_create_url_variable(model_factory):
    var = model_factory.create_url_variable('aname', 'http://someurl')

    assert var.id == 'aname'
    assert type(var.types[0]) == URLType

    value = var.initialValues[0].value
    assert type(value) == URL
    assert value.url == 'http://someurl'


def test_create_text_variable(model_factory):
    var = model_factory.create_text_variable('aname', 'a text')

    assert var.id == 'aname'
    assert type(var.types[0]) == TextType

    value = var.initialValues[0].value
    assert type(value) == Text
    assert value.text == 'a text'


def test_create_point_variable(model_factory):
    var = model_factory.create_point_variable('aname', 1.0, 2.0, 3.0)

    assert var.id == 'aname'
    assert type(var.types[0]) == PointType

    value = var.initialValues[0].value
    assert type(value) == Point
    assert value.x == 1.0
    assert value.y == 2.0
    assert value.z == 3.0


def test_create_expression_variable(model_factory):
    var = model_factory.create_expression_variable('aname', '')

    assert var.id == 'aname'
    assert type(var.types[0]) == ExpressionType

    value = var.initialValues[0].value
    assert type(value) == Expression


def test_create_visual_variable(model_factory):
    var = model_factory.create_visual_variable('aname')

    assert var.id == 'aname'
    assert type(var.types[0]) == VisualType


def test_create_image_variable(model_factory):
    var = model_factory.create_image_variable('aname')

    assert var.id == 'aname'
    assert type(var.types[0]) == ImageType


def test_create_connection_variable(model_factory):
    var = model_factory.create_connection_variable('aname')

    assert var.id == 'aname'
    assert type(var.types[0]) == ConnectionType


def test_create_particles_variable(model_factory):
    var = model_factory.create_particles_variable('aname', [])

    assert var.id == 'aname'
    assert type(var.types[0]) == VisualType


def test_create_pointer_variable(model_factory):
    var = model_factory.create_pointer_variable('aname', Pointer())

    assert var.id == 'aname'
    assert type(var.types[0]) == PointerType


def test_create_simple_array_variable(model_factory):
    var = model_factory.create_simple_array_variable('aname', IntArray([1, 2]))

    assert var.id == 'aname'
    assert type(var.types[0]) == SimpleArrayType
    value = var.initialValues[0].value
    assert type(value) == IntArray
    assert value.elements[0] == 1


def test_create_metadata_variable(model_factory):
    metadata = Metadata(
        [StringToValueMap('akey', Text('avalue')), StringToValueMap('anotherkey', Text('anothervalue'))])

    var = model_factory.create_metadata_variable('aname', metadata)

    assert var.id == 'aname'
    assert type(var.types[0]) == MetadataType
    value = var.initialValues[0].value
    assert type(value) == Metadata


def test_create_node_instance(model_factory):
    value = JSON('{"a": 1}')
    instance = model_factory.create_node_instance(id='inst', value=value)
    assert json.loads(instance.value.json)['a'] == 1


def test_create_edge_instance(model_factory):
    a = model_factory.create_node_instance(id='a', value=JSON('{"a": 1}'))
    b = model_factory.create_node_instance(id='b', value=JSON('{"b": 1}'))
    edge = model_factory.create_edge_instance('ab', a, b, Connectivity.from_string('DIRECTIONAL'),
                                              value=Metadata([StringToValueMap('akey', Text('avalue')),
                                                              StringToValueMap('anotherkey', Text('anothervalue'))]
                                                             )
                                              )
    assert json.loads(edge.a.value.json)['a'] == 1
    assert json.loads(edge.b.value.json)['b'] == 1
    assert str(edge.connectivity) == 'DIRECTIONAL'
    assert type(edge.value) == Metadata



import json

from pyecore.resources import ResourceSet, URI
from pygeppetto.data_model import GeppettoProject
from pygeppetto.model import GeppettoModel, GeppettoLibrary, CompositeType, Variable, ProcessQuery, DataSource, \
    QueryResults, SimpleInstance
from pygeppetto.model.datasources import QueryResult
from pygeppetto.model.model_access import GeppettoModelAccess
from pygeppetto.model.model_factory import GeppettoModelFactory
from pygeppetto.model.types import ImportType
from pygeppetto.model.values import ImportValue
from pygeppetto.services.data_manager import GeppettoDataManager
from pygeppetto.services.data_source_service import QueryProcessor, DataSourceService
from pygeppetto.services.model_interpreter import ModelInterpreter
from tests.test_xmi import filepath


class MockModelInterpreter(ModelInterpreter):

    def __init__(self):
        self.factory = GeppettoModelFactory()

    def create_model(self, url=None, typeName='MyGeppettoModel', library=None, commonLibraryAccess=None):
        '''
        Returns a geppetto model with this structure:

        v1 -> StateVariable ImportValue
        v2 -> StateVariable TimeSeries
        v3 -> compositeType ct1
            v31 -> StateVariable ImportValue
            v32 -> StateVariable TimeSeries

        :param url:
        :param typeName:
        :param library:
        :param commonLibraryAccess:
        :return:
        '''

        model = GeppettoModel(id='typeName', name=typeName, libraries=[self.factory.geppetto_common_library, library])

        v1 = self.factory.create_time_series_variable(id='v1', values=[1, 2, 3], unit='s')
        v2 = self.factory.create_time_series_variable(id='v2', values=[1, 2, 3], unit='s')
        v3 = Variable(id='v3')
        model.variables.append(v1)
        model.variables.append(v2)
        model.variables.append(v3)

        v31 = self.factory.create_state_variable(id='v31', initialValue=ImportValue())
        v32 = self.factory.create_time_series_variable(id='v32', values=[1, 2, 3], unit='s')

        ct = CompositeType(name='ct1', id='ct1', variables=[v31, v32])

        library.types.append(ct)
        v3.types.append(ct)

        v4 = Variable(id="v4")
        no_autores_type = ImportType(id="v4", url='/whatever', autoresolve=False)
        library.types.append(no_autores_type)
        v4.types.append(no_autores_type)
        v5 = Variable(id="v5")
        autores_type = ImportType(id="v5", url='/whatever/again', autoresolve=True)
        v5.types.append(autores_type)
        library.types.append(autores_type)

        model.variables.append(v4)
        model.variables.append(v5)
        return model

    def importValue(self, importValue):
        return self.factory.create_time_series(values=[4, 5, 6], unit='s')

    def importType(self, url, typeName, library, common_library_access: GeppettoModelAccess):
        assert url is not None
        assert typeName
        assert library
        assert common_library_access
        vi1 = self.factory.create_text_variable(id='vi', text='imported!!!')

        ct = CompositeType(name='ct2', id='ct2', variables=[vi1])
        library.types.append(ct)
        return ct

    def downloadModel(self, pointer, format, aspectConfiguration):
        pass

    def getSupportedOutputs(self, pointer):
        pass

    def getName(self):
        pass

    def getDependentModels(self):
        pass


def model_with_instances_datasource() -> GeppettoModel:
    rset = ResourceSet()
    resource = rset.get_resource(URI(filepath('instances_test.xmi')))
    return resource.contents[0]


class MockQueryProcessor(QueryProcessor):

    def process(self, query: ProcessQuery, data_source: DataSource, variable, results: QueryResults,
                model_access: GeppettoModelAccess) -> QueryResults:
        print("Processing query {}".format(query.id))
        return results


class MockFetchQueryProcessor(QueryProcessor):
    variable_name = "set by MockFetchQueryProcessor"

    def process(self, query: ProcessQuery, data_source: DataSource, query_param, results: QueryResults,
                model_access: GeppettoModelAccess) -> QueryResults:
        print("Processing query {}".format(query.id))

        if "var" in query_param:
            variable = Variable(id=query_param)
            variable.name = self.variable_name
            variable.types.append(model_access.geppetto_common_library.types[0])
            model_access.add_variable(variable)

        else:
            instance = SimpleInstance(id=query_param)
            instance.type = model_access.geppetto_common_library.types[0]
            instance.name = self.variable_name
            model_access.add_instance(instance)
        return results


class MockDataSourceService(DataSourceService):
    def get_template(self):
        return '{"statement":"$QUERY"}'

    def get_connection_type(self):
        return 'POST'

    def process_response(self, response):
        response = json.loads(response[0])['results'][0]
        query_results = QueryResults()
        query_results.header.extend(response['columns'])
        query_results.results.extend(QueryResult(values=r) for r in response['data'])

        return query_results


class MockDataManager(GeppettoDataManager):

    def get_project_from_url(self, url=None):
        project = GeppettoProject(id='mock', name='mock', geppetto_model=model_with_instances_datasource())
        self.projects[project.id] = project
        return project


def neo4j_response():
    # QUERY: "\"statement\": \"MATCH(n) RETURN id(n) as ID, n;\""
    # url: "http://localhost:7474/db/data/transaction/commit"
    return {
        "results": [{
            "columns": ["ID", "n"],
            "data": [
                {
                    "row": [
                        0,
                        {
                            "title": "The Matrix",
                            "released": 1999
                        }
                    ],
                    "meta": [
                        None,
                        {
                            "id": 0,
                            "type": "node",
                            "deleted": False
                        }
                    ]
                },
                {
                    "row": [
                        1,
                        {
                            "released": 1964,
                            "title": "Keanu Reeves"
                        }
                    ],
                    "meta": [
                        None,
                        {
                            "id": 1,
                            "type": "node",
                            "deleted": False
                        }
                    ]
                }
            ]
        }],
        "errors": []
    }


def neo4j_response_error():
    # QUERY: "\"statement\": \"MATCH(n) RETXXXXXXURN (n);\""
    # url: "http://localhost:7474/db/data/transaction/commit"
    return [{
        "results": [],
        "errors": [{
            "code": "Neo.ClientError.Statement.SyntaxError",
            "message": "Invalid input 'X': expected 'e/E' (line 1, column 11 (offset: 10))\\n\\\"MATCH(n) RXTURN (n);\\\"\\n^"
        }]
    }]
